from eflow.utils.string_utils import convert_to_filename, \
    correct_directory_path
import os
import pickle
import json

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

def check_if_directory_exists(directory_path):
    """

        Checks if the given directory path exists. Raises an error if doesn't

    Args:
        directory_path: string
            Given path that already exists.

    Raises:
        Raise an error if the given directory does not exist on the users system.
    """
    if not os.path.exists(directory_path):
        raise SystemError("Main directory path doesn't exist.\n"
                          "To help ensure unwanted directories are created you "
                          "must have a pre-defined path.")

def create_unique_directory(directory_path,
                            folder_name):
    """

        Creates a unique folder in the proper directory structure.

    Args:
        directory_path: string
            Given path that already exists.

        folder_name: string
            Folder name to generated.
    """

    os.makedirs(get_unique_directory_path(directory_path,
                                          folder_name))

def get_unique_directory_path(directory_path,
                              folder_name):
    """

        Iterate through directory structure until a unique folder name can be
        found.

        Note:
            Keeps changing the folder name by appending 1 each iteration.

    Args:
        directory_path: string
            Given path that already exists.

        folder_name: string
             Folder name to compare against other directories that exist in the
             directory_path.

    Returns:
        Returns back a directory path with a unique folder name.
    """

    # -----
    directory_path = correct_directory_path(directory_path)
    check_if_directory_exists(directory_path)

    create_dir_structure(directory_path=directory_path,
                         create_sub_dir="")

    # Ensures the folder is unique in the directory
    iterable = 0
    while True:
        if iterable != 0:
            created_path = f'{directory_path}{folder_name} {iterable}'
        else:
            created_path = f'{directory_path}{folder_name}'

        if not os.path.exists(created_path):
            break

        iterable += 1

    return created_path


def create_dir_structure(directory_path,
                         create_sub_dir):
    """

        Creates required directory structures inside the parent
        directory figures.

    Args:
        directory_path: string
            Given path that already exists.

        create_sub_dir: string
            Sub directory to create a given folder path.

    Returns:
        Returns back the created directory.
    """
    directory_path = correct_directory_path(directory_path)
    check_if_directory_exists(directory_path)

    for directory in create_sub_dir.split("/"):
        directory_path += "/" + directory
        if not os.path.exists(directory_path):
            os.makedirs(directory_path)

    return correct_directory_path(directory_path)


def write_object_text_to_file(obj,
                              directory_path,
                              filename,
                              remove_file_extension=True):
    """

        Writes the object's string representation to a text file.

    Args:
        obj: any
            Any object that has a string 'repr'.

        directory_path: string
            Given path that already exists.

        filename: string
            Text file's name.
    """
    directory_path = correct_directory_path(directory_path)
    check_if_directory_exists(directory_path)

    filename = convert_to_filename(filename, remove_file_extension=remove_file_extension)
    file_dir = f'{directory_path}{filename}.txt'

    f = open(file_dir, 'w')
    f.write('obj = ' + repr(obj) + '\n')
    f.close()

def pickle_object_to_file(obj,
                          directory_path,
                          filename,
                          remove_file_extension=True):
    """

        Writes the object to a pickle file.

    Args:
        obj: any object
            Any python object that can be pickled.

        directory_path: string
            Given path that already exists.

        filename: string
             Pickle file's name.
    """
    try:
        directory_path = correct_directory_path(directory_path)
        check_if_directory_exists(directory_path)

        # Ensures no file extensions in filename
        filename = convert_to_filename(filename,
                                       remove_file_extension=remove_file_extension)
        file_dir = f'{directory_path}{filename}.pkl'
        list_pickle = open(file_dir, 'wb')
        pickle.dump(obj,
                    list_pickle)
    finally:
        list_pickle.close()

    return file_dir

def load_pickle_object(file_path):
    with open(file_path, 'rb') as handle:
        return pickle.load(handle)

def dict_to_json_file(dict_obj,
                      directory_path,
                      filename,
                      remove_file_extension=True):
    """

        Writes a dict to a json file.

    Args:
        dict_obj: dict
            Dictionary object.

        directory_path: string
            Given path that already exists.

        filename: string
            Json file's name.
    """
    directory_path = correct_directory_path(directory_path)
    check_if_directory_exists(directory_path)

    filename = convert_to_filename(filename,remove_file_extension=remove_file_extension)

    with open(f'{directory_path}{filename}.json',
              'w',
              encoding='utf-8') as outfile:
        json.dump(dict_obj,
                  outfile,
                  ensure_ascii=False,
                  indent=2)

def json_file_to_dict(filepath):
    """

        Returns back the dictionary from of a json file.

    Args:
        filepath: string
             Given path to the filename.

    Returns:
        Returns back the dictionary from of a json file.
    """
    json_file = open(filepath)
    json_str = json_file.read()
    json_data = json.loads(json_str)
    json_file.close()

    return json_data

def get_all_directories_from_path(directory_path):
    """

       Gets directories names with the provided path.

    Args:
        directory_path: string
            Given path that already exists.

    Returns:
        Returns back a set a directories with the provided path.
    """
    directory_path = correct_directory_path(directory_path)
    check_if_directory_exists(directory_path)

    dirs_in_paths = []
    for (dirpath, dirnames, filenames) in os.walk(directory_path):
        dirs_in_paths.extend(dirnames)
        break

    return set(dirs_in_paths)


def get_all_files_from_path(directory_path,
                            file_extension=None):
    """

        Gets all filenames with the provided path.

    Args:
        directory_path: string
            Given path that already exists.

        file_extension: string
            Only return files that have a given extension.

    Returns:
        Returns back a set a filenames with the provided path.
    """
    directory_path = correct_directory_path(directory_path)
    check_if_directory_exists(directory_path)

    files_in_paths = []
    for (dirpath, dirnames, filenames) in os.walk(directory_path):

        if file_extension:
            file_extension = file_extension.replace(".","")
            for file in filenames:
                if file.endswith(f'.{file_extension}'):
                    files_in_paths.append(file)
        else:
            files_in_paths.extend(filenames)
        break

    return set(files_in_paths)



from eflow._hidden.parent_objects import FileOutput
from eflow._hidden.custom_exceptions import PipelineSegmentError, UnsatisfiedRequirments
from eflow.utils.sys_utils import dict_to_json_file,json_file_to_dict,get_all_files_from_path, create_dir_structure
from eflow.utils.string_utils import create_hex_decimal_string

from collections import deque
import copy
import os
import inspect
import re

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class DataPipelineSegment(FileOutput):
    """
        Holds the function name's and arguments to be pushed to a json file.
        This generated json file later allows a dataframe to be passed to it
        and applies the functions and arguments in the sequence they were run.
    """
    def __init__(self,
                 object_type,
                 segment_id=None,
                 create_file=True):
        """
        Args:
            object_type: string
                The child type of all object's that inherited DataPipelineSegment

            segment_id: string
                 If init as a string instead of None; the object will attempt
                 to find the json file in the provided directory.
        Note:
            Essentially we are serializing the object with json files.
        """

        self.__json_file_name = None
        self.__object_type = copy.deepcopy(object_type)

        if not isinstance(segment_id, str) and segment_id:
            raise UnsatisfiedRequirments(
                "Segment id must be a string or set to 'None'!")

        if segment_id and not create_file:
            raise PipelineSegmentError("Parameter conflict: segment_id is referring "
                                       "to a saved file but create_file is set to False.")

        # File extension removal
        if isinstance(segment_id,str):
            segment_id = segment_id.split(".")[0]
        self.__segment_id = copy.deepcopy(segment_id)

        # Pushes the functions info based on order they are called
        self.__function_pipe = deque()

        self.__create_file = create_file
        self.__lock_interaction = False

        # Attempt to get json file into object's attributes.
        if self.__segment_id:
            self.__configure_pipeline_segment_with_existing_file()


    def perform_segment(self,
                        df,
                        df_features):
        """

            Performs all functions that the child of the pipeline segment has
            performed on a given pandas dataframe.

        Args:
            df: pd.Dataframe
                Pandas Dataframe

            df_features: DataFrameTypes from eflow
                DataFrameTypes object.
        """
        for function_name, params_dict in self.__function_pipe:

            # Get function
            method_to_call = getattr(self, function_name)

            # -----
            params_dict["df"] = df
            params_dict["df_features"] = df_features
            params_dict["_add_to_que"] = False

            # Function call with arguments
            method_to_call(**params_dict)

            # -----
            del params_dict["df"]
            del params_dict["df_features"]
            del params_dict["_add_to_que"]

    # def generate_code(self,
    #                   generate_file=True,
    #                   add_libs=True):
    #     """
    #
    #         Attempts to parse the file of the child object of the pipline
    #         segment.
    #
    #     Args:
    #         generate_file:
    #             Depending on the boolean value of generate_file; True will
    #             generate a python file and False will just return a list of
    #             strings.
    #
    #         add_libs:
    #             Adds utils libs to the top of list.
    #
    #     Returns:
    #         If the arg 'generate_file' is set to False then it will just return
    #         a list of strings.
    #     """
    #
    #     # Raise error if no methods of the child have been called yet
    #     if len(self.__function_pipe) == 0:
    #         raise PipelineSegmentError(
    #             "Can't generate code when no methods of this segment have been used yet!")
    #
    #     generated_code = []
    #
    #     # Add libs to the top of the file
    #     if add_libs:
    #         generated_code.append("from eflow.utils.math_utils import *")
    #         generated_code.append("from eflow.utils.image_processing_utils import *")
    #         generated_code.append("from eflow.utils.pandas_utils import *")
    #         generated_code.append("from eflow.utils.modeling_utils import *")
    #         generated_code.append("from eflow.utils.string_utils import *")
    #         generated_code.append("from eflow.utils.misc_utils import *")
    #         generated_code.append("from eflow.utils.sys_utils import *")
    #         generated_code.append("")
    #
    #     # -----
    #     for function_name, params_dict in self.__function_pipe:
    #
    #         # Get function's code
    #         pre_made_code = inspect.getsource(getattr(self, function_name))
    #
    #         first_lines_found = False
    #         def_start = False
    #
    #         # Init variables
    #         for parm, val in params_dict.items():
    #             generated_code.append(f"{parm} = {val}")
    #
    #         # Iterate line by line of function's code
    #         for line in pre_made_code.split("\n"):
    #
    #             if "def " in line:
    #                 def_start = True
    #                 continue
    #
    #             if def_start:
    #                 if "):" in line:
    #                     def_start = False
    #                 continue
    #
    #             if "params_dict" in line or "_add_to_que" in line:
    #                 continue
    #
    #             if not re.search('[a-zA-Z]', line) and not first_lines_found:
    #                 continue
    #             else:
    #                 first_lines_found = True
    #
    #             if "__add_function_to_que" in line:
    #                 continue
    #
    #             # -----
    #             generated_code.append(line.replace("        ", "", 1))
    #
    #         # Formatting
    #         generated_code.append("# " + "------" * 5)
    #
    #     # Generate file or pass back list
    #     if generate_file:
    #         create_dir_structure(self.folder_path,
    #                                    "Generated code")
    #         with open(
    #                 self.folder_path + f'Generated code/{self.__segment_id}.py',
    #                 'r+') as filehandle:
    #             filehandle.truncate(0)
    #             for listitem in generated_code:
    #                 filehandle.write('%s\n' % listitem)
    #         print(
    #             f"Generated a python file named/at: {self.folder_path}Generated code/{self.__segment_id}.py")
    #     else:
    #         return generated_code

    def reset_segment_file(self):
        # File/Folder error checks
        if not os.path.exists(self.folder_path):
            raise PipelineSegmentError(
                "Couldn't find the pipeline segment's folder when trying to configure this object with the provided json file.")
        if not os.path.exists(
                self.folder_path + copy.deepcopy(self.__json_file_name)):
            raise PipelineSegmentError(
                f"Couldn't find the pipeline segment's file named '{self.__json_file_name}' in the pipeline's directory when trying to configure this object with the provided json file.")

        dict_to_json_file({},
                          self.folder_path,
                          self.file_name)

    @property
    def file_path(self):
        """

            Gets the absolute path to the json file.
        """
        # -----
        if len(self.__function_pipe) == 0:
            raise PipelineSegmentError("The pipeline segment has not performed any actions yet."
                                       " Please perform some methods with this object.")
        elif not self.__create_file:
            raise PipelineSegmentError("This pipeline segment does not have saved "
                                       "file and thus can not have a file path.")
        else:
            return self.folder_path + copy.deepcopy(self.__json_file_name)

    @property
    def file_name(self):
        """

            Gets the file name of the json file.
        """
        # -----
        if len(self.__function_pipe) == 0:
            raise PipelineSegmentError("The pipeline segment has not performed any actions yet."
                                       " Please perform some methods with this object.")
        elif not self.__create_file:
            raise PipelineSegmentError("This pipeline segment does not have saved "
                                       "file and thus can not have a file path.")
        else:
            return copy.deepcopy(self.__json_file_name)


    def __replace_function_in_que(self,
                                  function_name,
                                  params_dict,
                                  param,
                                  param_val):

        raise ValueError("This function hasn't been completed yet!")

        if self.__lock_interaction:
            raise PipelineSegmentError("This pipeline has be locked down and "
                                       "will prevent futher changes to the generated flat file.")

        for delete_key in {"self", "df", "df_features", "_add_to_que",
                           "params_dict"}:
            if delete_key in params_dict.keys():
                del params_dict[delete_key]

        for k, v in {k: v for k, v in params_dict.items()}.items():
            if k not in parameters:
                del params_dict[k]
            elif isinstance(v, set):
                params_dict[k] = list(v)

        self.__function_pipe.append((function_name,
                                     params_dict))

        # Generate new json file name with proper file/folder output attributes
        if len(self.__function_pipe) == 1 and not self.__json_file_name:
            FileOutput.__init__(self,
                                f'_Extras/Pipeline Structure/Data Pipeline Segments/{self.__object_type}')
            all_json_files = get_all_files_from_path(self.folder_path,
                                                     ".json")
            while True:
                random_file_name = create_hex_decimal_string().upper()
                if random_file_name not in all_json_files:
                    break

            self.__segment_id = random_file_name
            self.__json_file_name = random_file_name + ".json"

        # Update json file
        if self.__create_file:
            self.__create_json_pipeline_segment_file()


    def __add_function_to_que(self,
                              function_name,
                              parameters,
                              params_dict):
        """

            Adds the function info the function que. If the segment has no
            json file name then generate one for it the given directory.

        Args:
            function_name: string
                Functions name

            params_dict: dict
                Parameter's name to their associated values.

        Note:
            This function should only ever be called by children of
            this object.
        """
        if self.__lock_interaction:
            raise PipelineSegmentError("This pipeline has be locked down and "
                                       "will prevent futher changes to the generated flat file.")


        for delete_key in {"self", "df", "df_features", "_add_to_que",
                    "params_dict"}:
            if delete_key in params_dict.keys():
                del params_dict[delete_key]


        for k,v in {k:v for k,v in params_dict.items()}.items():
            if k not in parameters:
                del params_dict[k]
            elif isinstance(v,set):
                params_dict[k] = list(v)

        self.__function_pipe.append((function_name,
                                     params_dict))

        # Generate new json file name with proper file/folder output attributes
        if len(self.__function_pipe) == 1 and not self.__json_file_name:
            FileOutput.__init__(self,
                                f'_Extras/Pipeline Structure/Data Pipeline Segments/{self.__object_type}')
            all_json_files = get_all_files_from_path(self.folder_path,
                                                     ".json")
            while True:
                random_file_name = create_hex_decimal_string().upper()
                if random_file_name not in all_json_files:
                    break

            self.__segment_id = random_file_name
            self.__json_file_name = random_file_name + ".json"

        # Update json file
        if self.__create_file:
            self.__create_json_pipeline_segment_file()

    def __create_json_pipeline_segment_file(self):
        """

            Creates a json file that the segment relates to; this file's
            segment id can be used later used in the init to get all functions
            that object already used.
        """
        json_dict = dict()

        # Pipeline Segment
        json_dict["Pipeline Segment"] = dict()
        json_dict["Pipeline Segment"]["Object Type"] = self.__object_type
        json_dict["Pipeline Segment"]["Functions Performed Order"] = dict()
        function_order = 1
        for function_name, params_dict in self.__function_pipe:
            json_dict["Pipeline Segment"]["Functions Performed Order"][f"Function " \
                f"Order {function_order}"] = dict()
            json_dict["Pipeline Segment"]["Functions Performed Order"][f"Function " \
                f"Order {function_order}"][function_name] = dict()

            json_dict["Pipeline Segment"]["Functions Performed Order"][
                f"Function Order {function_order}"][function_name][
                "Params Dict"] = params_dict

            function_order += 1
        json_dict["Pipeline Segment"]["Function Count"] = function_order - 1

        # Generate pipeline segment file
        dict_to_json_file(json_dict,
                          self.folder_path,
                          self.__json_file_name)

    def __configure_pipeline_segment_with_existing_file(self):
        """

            Attempts to get a json file and then re_init the 'function_pipe'
            and the 'json_file_name'.
        """

        FileOutput.__init__(self,
                            f'_Extras/Pipeline Structure/Data Pipeline Segments/{self.__object_type}')

        self.__function_pipe = deque()
        self.__json_file_name = copy.deepcopy(self.__segment_id) + ".json"

        # File/Folder error checks
        if not os.path.exists(self.folder_path):
            raise PipelineSegmentError(
                "Couldn't find the pipeline segment's folder when trying to configure this object with the provided json file.")
        if not os.path.exists(self.folder_path + copy.deepcopy(self.__json_file_name)):
            raise PipelineSegmentError(
                f"Couldn't find the pipeline segment's file named '{self.__json_file_name}' in the pipeline's directory when trying to configure this object with the provided json file.")

        json_dict = json_file_to_dict(
            self.folder_path + copy.deepcopy(self.__json_file_name))

        # Push functions into function pipe
        for function_order in range(1,json_dict["Pipeline Segment"]["Function Count"] + 1):
            function_name = list(json_dict["Pipeline Segment"]["Functions Performed Order"][f"Function Order {function_order}"].keys())[0]
            params_dict = json_dict["Pipeline Segment"]["Functions Performed Order"][f"Function Order {function_order}"][function_name]["Params Dict"]
            self.__function_pipe.append((function_name,
                                         params_dict))




from eflow._hidden.parent_objects import FileOutput
from eflow._hidden.parent_objects import DataPipelineSegment
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments, PipelineError
from eflow._hidden.constants import SYS_CONSTANTS
from eflow.utils.string_utils import create_hex_decimal_string,correct_directory_path
from eflow.utils.sys_utils import dict_to_json_file, create_dir_structure, json_file_to_dict
from eflow.utils.eflow_utils import move_folder_to_eflow_garbage
from eflow.data_pipeline_segments import *

from eflow.foundation import DataFrameTypes

import copy
import os
from collections import deque

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class DataPipeline(FileOutput):
    """
        Houses all pipeline segments. Able to load these segments in with
        a pipeline json file and/or modify id. You can add more pipeline
        segments if needed which will modify the pipeline's json file. Allows
        a pandas dataframe to be processed by 'perform_pipeline'.
    """
    def __init__(self,
                 pipeline_name,
                 df,
                 df_features=None,
                 pipeline_modify_id=None,
                 remove_past_contents=False):
        """
        Args:
            pipeline_name (str):
                Points to/generates a folder based on the pipeline's name.

            pipeline_modify_id (str,NoneType):
                If set to 'None' then will point the 'root' or the main template
                of the pipeline.

            remove_past_contents:
                If an already existing folder exists for this then move to
                eflow's personal garbage.
        """
        # Set up directory structure
        dir_path_to_pipeline = correct_directory_path(f"{os.getcwd()}/{SYS_CONSTANTS.PARENT_OUTPUT_FOLDER_NAME}/_Extras/Pipeline Structure/Data Pipeline/{pipeline_name}")
        configure_existing_file = False

        # Get json proper file name
        if pipeline_modify_id:
            json_file = f"{pipeline_modify_id.split('.')[0]}.json"
        else:
            json_file = "root_pipeline.json"

        self.__df_features = None

        # Check if folder/file exist for the pipeline
        if os.path.exists(dir_path_to_pipeline):
            if os.path.exists(dir_path_to_pipeline + json_file):
                print(f"The file '{json_file}' exist!")
                FileOutput.__init__(self,
                                    f'_Extras/Pipeline Structure/Data Pipeline/{pipeline_name}')
                configure_existing_file = True
            else:
                raise PipelineError(f"The file '{json_file}' does not exist!")

            # Create/Load in df_features to given object.
            if os.path.exists(dir_path_to_pipeline + "df_features.json"):
                df_features = DataFrameTypes(None)
                df_features.init_on_json_file(dir_path_to_pipeline + "df_features.json")
            else:
                if df_features is None:
                    raise PipelineError("When initializing a data pipeline structure "
                                        "you must pass a DataFrameTypes object with "
                                        "the correctly defined types!")

                # Create file representation
                else:
                    df_features.create_json_file_representation(dir_path_to_pipeline,
                                                                "df_features")

        # -----
        if df_features is None:
            raise PipelineError("When initializing a data pipeline structure "
                                "you must pass a DataFrameTypes object with "
                                "the correctly defined types!")
        self.__df_features = copy.deepcopy(df_features)

        # Check if root file exist or if pipeline modify id
        self.__pipeline_name = copy.deepcopy(pipeline_name)
        self.__pipeline_segment_deque = deque()
        self.__pipeline_segment_names = set()
        self.__pipeline_segment_path_id = set()
        self.__pipeline_modify_id = copy.deepcopy(pipeline_modify_id)

        self.__json_file_name = json_file

        # Json file does exist; init DataPipeline object correctly
        if configure_existing_file:
            if remove_past_contents:
                print("Moving past contents to eFlow's garbage.")
                move_folder_to_eflow_garbage(dir_path_to_pipeline,
                                             "Data Pipelines")
            else:
                print("Now configuring object with proper pipeline segments...")
                self.__configure_pipeline_with_existing_file()

        for removal_feature in set(df.columns) ^ set(
                df_features.all_features()):
            print(f"Removing the feature: \"{removal_feature}\"")
            df.drop(columns=removal_feature,
                    inplace=True)

    @property
    def file_path(self):
        """
        Returns:
            File path with file name.
        """
        if len(self.__pipeline_segment_deque) == 0:
            raise PipelineError(
                "The pipeline has not added any segments yet."
                " Please add segments to this object.")
        else:
            return self.folder_path + copy.deepcopy(self.__json_file_name)

    @property
    def file_name(self):
        """
        Returns:
            File name with extension.
        """
        if len(self.__pipeline_segment_deque) == 0:
            raise PipelineError(
                "The pipeline has not added any segments yet."
                " Please add segments to this object.")
        else:
            return copy.deepcopy(self.__json_file_name)

    def get_df_features(self):
        if len(self.__pipeline_segment_deque) == 0:
            raise PipelineError(
                "The pipeline has not added any segments yet."
                " Please add segments to this object.")
        else:
            return copy.deepcopy(self.__df_features)

    def add(self,
            segment_name,
            pipeline_segment_obj):
        """
        segment_name (str):
            A aliased name to refer to this segment.

        pipeline_segment_obj (child of DataPipelineSegment):
            A child object of type DataPipelineSegment.

        Returns:
            Attempts to add a pipeline segment object to the objects que and
            update it's related json object.
        """

        # pipeline_segment_obj = copy.deepcopy(pipeline_segment_obj)

        # Type check
        if not isinstance(pipeline_segment_obj,
                          DataPipelineSegment):

            raise UnsatisfiedRequirments(f"Expected a 'DataPipelineSegment' object; received '{type(pipeline_segment_obj)}'")

        # Check if alias has already been used
        if segment_name in self.__pipeline_segment_names:
            raise PipelineError(f"The '{segment_name}' pipeline segment is already in this pipeline. Please choose a different segment name.")

        try:
            # Check if the pipeline segment has already been used
            segment_path_id = pipeline_segment_obj.relative_folder_path + pipeline_segment_obj.file_name
        except AttributeError:
            raise UnsatisfiedRequirments("The given pipeline segment didn't perform any functions.")

        if segment_path_id in self.__pipeline_segment_path_id:
            raise PipelineError("The segment has been already found "
                                "in this pipeline Segment path id: " +
                                f"'{segment_path_id}.'\n" +
                                "This can be done by:"
                                "\n\t*Creating a completely new segment object "
                                "and adding it to the pipeline with the 'add'"
                                " method."
                                "\n\t*Refer to a different segment path id")
        else:
            # Que has yet to have data pushed; set up output directory for new data
            if len(self.__pipeline_segment_deque) == 0:
                FileOutput.__init__(self,
                                    f'_Extras/Pipeline Structure/Data Pipeline/{self.__pipeline_name}')

                if os.path.exists(self.folder_path + "df_features.json"):
                    self.__df_features.init_on_json_file(self.folder_path + "df_features.json")
                else:
                    self.__df_features.create_json_file_representation(self.folder_path,
                                                                       "df_features.json")



        # Update data types for error checking
        self.__pipeline_segment_names.add(segment_name)
        self.__pipeline_segment_path_id.add(segment_path_id)

        # Main component of the project
        self.__pipeline_segment_deque.append((segment_name,
                                              segment_path_id,
                                              pipeline_segment_obj))

        # Lock down the object to prevent users from continuing to interact with it after adding it to the pipeline
        pipeline_segment_obj._DataPipelineSegment__lock_interaction = True

        # Update/Create the json file
        self.__create_json_pipeline_file()

    def perform_pipeline(self,
                         df,
                         df_features=None):
        """
        Args:
            df:
                Pandas Dataframe object to be transformed by the pipeline.

        Returns:
            Applies a Pandas Dataframe object to all functions on all segments
            in the pipeline.
        """

        if df_features is None:
            df_features = self.__df_features

        if self.__df_features is None:
            raise PipelineError("Default type holder somehow is equal to none for "
                                "this pipeline structure.")


        for removal_feature in set(df.columns) ^ set(
                df_features.all_features()):
            print(f"Removing the feature: \"{removal_feature}\"")
            df.drop(columns=removal_feature,
                    inplace=True)

        for _, _, pipeline_segment in self.__pipeline_segment_deque:
            pipeline_segment.perform_segment(df,
                                             df_features)

    def __create_json_pipeline_file(self):
        """
        Returns:
            Creates a dict based on the given contents of the variable
            'self.__pipeline_segment_deque' to convert to a json file.
            This file will later be used to instruct our object to execute
            specific code.
        """

        # -------------
        json_dict = dict()
        segment_order = 1

        json_dict["Pipeline Name"] = self.__pipeline_name
        json_dict["Pipeline Segment Order"] = dict()
        for segment_name, segment_path_id, pipeline_segment_obj in self.__pipeline_segment_deque:
            json_dict["Pipeline Segment Order"][segment_order] = dict()
            json_dict["Pipeline Segment Order"][segment_order]["Pipeline Segment Path"] = segment_path_id
            json_dict["Pipeline Segment Order"][segment_order]["Pipeline Segment Type"] = pipeline_segment_obj.__class__.__name__
            json_dict["Pipeline Segment Order"][segment_order]["Pipeline Segment Name"] = segment_name
            json_dict["Pipeline Segment Order"][segment_order]["Pipeline Segment ID"] = segment_path_id.split("/")[-1].split(".")[0]

            segment_order += 1

        json_dict["Pipeline Segment Count"] = segment_order - 1

        # Create a folder for all non-root json files.
        if self.__pipeline_modify_id:
            create_dir_structure(self.folder_path,
                                       "/Modified Pipelines")
            dict_to_json_file(json_dict,
                                       self.folder_path + "/Modified Pipelines",
                                       self.__json_file_name)
        # Root json files only
        else:
            dict_to_json_file(json_dict,
                                       self.folder_path,
                                       self.__json_file_name)

    def __configure_pipeline_with_existing_file(self):
        """
        Returns:
            Changes the objects variables based on the provided json file.
        """

        # Error check paths
        if not os.path.exists(self.folder_path):
            raise PipelineError("Couldn't find the pipeline's folder when trying to configure this object with the provided json file.")

        # ------
        if not os.path.exists(self.folder_path + copy.deepcopy(self.__json_file_name)):
            raise PipelineError(f"Couldn't find the pipeline's file named '{self.file_name}' in the pipeline's directory when trying to configure this object with the provided json file.")

        # Reset variables
        self.__pipeline_segment_deque = deque()
        self.__pipeline_segment_names = set()
        self.__pipeline_segment_path_id = set()

        json_dict = json_file_to_dict(self.folder_path + copy.deepcopy(self.__json_file_name))

        # Iterate through dict to init variables properly
        for segment_order in range(1,json_dict["Pipeline Segment Count"] + 1):
            segment_type = json_dict["Pipeline Segment Order"][str(segment_order)]["Pipeline Segment Type"]
            segment_name = \
            json_dict["Pipeline Segment Order"][str(segment_order)][
                "Pipeline Segment Name"]
            segment_path_id = \
            json_dict["Pipeline Segment Order"][str(segment_order)][
                "Pipeline Segment Path"]
            segment_id = json_dict["Pipeline Segment Order"][str(segment_order)]["Pipeline Segment ID"]

            pipeline_segment_obj = None

            pipeline_segment_obj = eval(f"{segment_type}(\"{segment_id}\")\n")

            if not pipeline_segment_obj:
                raise PipelineError(f"An unknown error has occurred with finding the correct pipeline segment for '{segment_type}' segment!")

            self.__pipeline_segment_names.add(segment_name)
            self.__pipeline_segment_path_id.add(segment_path_id)

            self.__pipeline_segment_deque.append((segment_name,
                                                  segment_path_id,
                                                  pipeline_segment_obj))

    # def generate_code(self,
    #                   generate_file=True,
    #                   add_libs=True):
    #     """
    #     generate_file (bool):
    #         Determines whether or not to generate a python file
    #     add_libs:
    #
    #     Returns:
    #         Applies
    #     """
    #     if len(self.__pipeline_segment_deque) == 0:
    #         raise PipelineError("Pipeline needs a segment to generate code.")
    #
    #     generated_code = []
    #
    #     if add_libs:
    #         libs_code = list()
    #         libs_code.append("from eflow.utils.math_utils import *")
    #         libs_code.append("from eflow.utils.image_processing_utils import *")
    #         libs_code.append("from eflow.utils.pandas_utils import *")
    #         libs_code.append("from eflow.utils.modeling_utils import *")
    #         libs_code.append("from eflow.utils.string_utils import *")
    #         libs_code.append("from eflow.utils.misc_utils import *")
    #         libs_code.append("from eflow.utils.sys_utils import *")
    #         libs_code.append("")
    #         generated_code.append(libs_code)
    #
    #
    #     for segment_name, segment_path_id, pipeline_segment in self.__pipeline_segment_deque:
    #         tmp_list = pipeline_segment.generate_code(generate_file=False,
    #                                                   add_libs=False)
    #         tmp_list.insert(0, f'# Segment Name:    {segment_name}')
    #         tmp_list.insert(0, f'# Segment Path ID: {segment_path_id}')
    #
    #         generated_code.append(tmp_list)
    #
    #     create_dir_structure(self.folder_path,
    #                                "Generated code")
    #
    #     if generate_file:
    #         python_file_name = self.__json_file_name.split(".")[0] + ".py"
    #         with open(self.folder_path +
    #                   f'Generated code/{python_file_name}',
    #                   'w') as filehandle:
    #             for segment_code in generated_code:
    #                 for listitem in segment_code:
    #                     filehandle.write('%s\n' % listitem)
    #         print(
    #             f"Generated a python file named/at: {self.folder_path}Generated code/{python_file_name}")
    #     else:
    #         return generated_code
    #



import multiprocessing as mp

def multiprocess(func,
                 jobs,
                 cores):
    results = []

    if cores == 1:
        for j in jobs:
            results.append(func(j))
    elif cores == -1:
        with mp.Pool(mp.cpu_count()) as p:
            results = list(p.map(func, jobs))
    elif cores > 1:
        with mp.Pool(cores) as p:
            results = list(p.map(func, jobs))
    else:
        print('Error: cores must be a integer')

    return results


from eflow.utils.image_processing_utils import create_plt_png
from eflow.utils.string_utils import correct_directory_path
from eflow.utils.sys_utils import write_object_text_to_file, create_dir_structure, pickle_object_to_file
from eflow._hidden.custom_exceptions import UnsatisfiedRequirments
from eflow.utils.math_utils import calculate_entropy

import pandas as pd
from matplotlib import pyplot as plt
import copy
import six
import numpy as np

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


def df_to_image(df,
                directory_path,
                sub_dir,
                filename,
                sharpness=1.7,
                col_width=8,
                row_height=0.625,
                font_size=14,
                header_color='#40466e',
                row_colors=['#f1f1f2', 'w'],
                edge_color='w',
                bbox=[0, 0, 1, 1],
                header_columns=0,
                ax=None,
                show_index=False,
                index_color="#add8e6",
                format_float_pos=None,
                show_plot=False,
                **kwargs):

    directory_path = correct_directory_path(directory_path)
    df = copy.deepcopy(df)

    if format_float_pos and format_float_pos >= 1:
        float_format = '{:,.' + str(format_float_pos) + 'f}'
        for col_feature in set(df.select_dtypes(include=["float"]).columns):
            df[col_feature] = df[col_feature].map(float_format.format)

    if ax is None:
        size = (np.array(df.shape[::-1]) + np.array([0, 1])) * np.array([col_width, row_height])
        fig, ax = plt.subplots(figsize=size)
        ax.axis('off')

    if show_index:
        df.reset_index(inplace=True)

    mpl_table = ax.table(cellText=df.values,
                         bbox=bbox,
                         colLabels=df.columns,
                         **kwargs)

    mpl_table.auto_set_font_size(False)
    mpl_table.set_fontsize(font_size)

    for k, cell in six.iteritems(mpl_table._cells):
        cell.set_edgecolor(edge_color)
        if k[0] == 0 or k[1] < header_columns:
            cell.set_text_props(weight='bold', color='w')
            cell.set_facecolor(header_color)
        else:
            if index_color and show_index and k[1] == 0:
                cell.set_facecolor(index_color)
            else:
                cell.set_facecolor(row_colors[k[0] % len(row_colors)])

    if not sub_dir:
        sub_dir = ""

    create_plt_png(directory_path,
                   sub_dir,
                   filename,
                   sharpness)
    if show_plot:
        plt.show()

    plt.close("all")


def average_feature_correlation_table(df):
    """

        Creates a correlation table of the feature's relationship base on the
        average correlation with other features.

    Args:
        df: pd.Dataframe
            Pandas dataframe

    Returns:
        A dataframe that shows the average correlations between each feature.
    """

    corr_metrics = df.corr()
    for index, feature_index in enumerate(corr_metrics.index.tolist()):
        corr_metrics.loc[feature_index][index] = np.nan

    corr_feature_means = dict()
    for feature_name in corr_metrics.columns:
        corr_feature_means[feature_name] = corr_metrics[
            feature_name].dropna().mean()

    corr_feature_means = pd.DataFrame.from_dict(corr_feature_means,
                                                orient='index',
                                                columns=["Average Correlations"])
    corr_feature_means.index.name = "Features"

    return corr_feature_means


def feature_correlation_table(df):
    """

        Creates a correlation table of each feature's relationship with one
        another.

    Args:
        df: pd.Dataframe
            Pandas dataframe

    Returns:
        A dataframe that shows the correlations between each feature.
    """
    feature_corr_dict = dict()
    df_corr = df.corr()

    remaining_features = df_corr.columns.tolist()
    for main_feature in df_corr.columns:
        for sub_feature in remaining_features:
            if main_feature != sub_feature:
                feature_corr_dict[f"{main_feature} to {sub_feature}"] = \
                df_corr[main_feature].loc[sub_feature]
        remaining_features.remove(main_feature)

    df_corr = pd.DataFrame.from_dict(feature_corr_dict,
                                     orient='index',
                                     columns=["Correlations"])
    df_corr.index.name = "Features"

    return df_corr


def check_if_feature_exists(df,
                            feature_name):
    """

        Checks if feature exists in the dataframe.

    Args:
        df:
            Pandas dataframe object.

        feature_name:
            Feature's name.

    Raises:
        Raises an error in the feature does not exist in the columns.
    """
    if feature_name not in df.columns:
        raise KeyError(
            f"The feature \'{feature_name}\' was not found in the dataframe!"
            + " Please select a valid feature from the dataframe")


def data_types_table(df,
                     sort_by_type=True):
    """

        Creates a pandas dataframe based on the features and their types of the
        given/passed dataframe.

    Args:
        df:
            Pandas DataFrame object

        sort_by_type:
            Orders the resulting dataframe by alphabetically by type.

    Returns:
        Returns a pandas dataframe of features and their found types
        in the passed dataframe.
    """

    if not df.shape[0]:
        print("Empty dataframe found!")
        return None

    dtypes_df = pd.DataFrame({'Data Types': df.dtypes.values})
    dtypes_df.index = df.dtypes.index.tolist()
    dtypes_df.index.name = "Features"
    dtypes_df["Data Types"] = dtypes_df["Data Types"].astype(str)

    if sort_by_type:
        dtypes_df = dtypes_df.sort_values("Data Types")

    return dtypes_df


def missing_values_table(df):
    """

        Creates a pandas dataframe based on the missing data inside the
        given/passed dataframe

    Args:
        df:
            Pandas DataFrame object

    Returns:
        Returns a Pandas DataFrame object giving the percentage of
        the null data for the original DataFrame columns.
    """

    mis_val = df.isnull().sum()
    mis_val_percent = 100 * df.isnull().sum() / len(df)
    mis_val_table = pd.concat([mis_val, mis_val_percent], axis=1)
    mis_val_table_ren_columns = mis_val_table.rename(
        columns={0: 'Missing Values', 1: '% of Total Values'})
    mis_val_table_ren_columns = mis_val_table_ren_columns[
        mis_val_table_ren_columns.iloc[:, 1] != 0].sort_values(
        '% of Total Values', ascending=False).round(1)

    mis_val_table_ren_columns.index.name = "Features"

    return mis_val_table_ren_columns


def generate_meta_data(df,
                       output_folder_path,
                       sub_dir):
    """

        Creates files representing the shape and feature types of the dataframe.

    Args:
        df: pd.Dataframe
            Pandas DataFrame object

        output_folder_path: str
            Pre defined path to already existing directory to output file(s).

        sub_dir: str
            Path to be possibly generated.

    Returns:
        Creates meta data on the passed datafrane.
    """
    create_dir_structure(output_folder_path,
                         correct_directory_path(sub_dir + "/Meta Data"))

    output_folder_path = correct_directory_path(output_folder_path)

    # Create files relating to dataframe's shape
    shape_df = pd.DataFrame.from_dict({'Rows': [df.shape[0]],
                                       'Columns': [df.shape[1]]})

    if shape_df.shape[0]:
        df_to_image(shape_df,
                    f"{output_folder_path}/{sub_dir}",
                    "Meta Data",
                    "Dataframe Shape Table",
                    show_index=False)

    write_object_text_to_file(shape_df.to_dict('records'),
                              f"{output_folder_path}/{sub_dir}/Meta Data",
                              "Dataframe Shape Text")

    # Create files relating to dataframe's types
    dtypes_df = data_types_table(df)
    if dtypes_df.shape[0]:
        df_to_image(dtypes_df,
                    f"{output_folder_path}/{sub_dir}",
                    "Meta Data",
                    "Dataframe Types Table",
                    show_index=True)

    plt.close("all")


    # Missing value table
    mis_val_table = missing_values_table(df)
    if mis_val_table.shape[0]:
        df_to_image(mis_val_table,
                    f"{output_folder_path}/{sub_dir}",
                    "Meta Data",
                    "Missing Data Table",
                    show_index=True)

    plt.close("all")

def generate_entropy_table(df,
                           df_features,
                           output_folder_path,
                           sub_dir,
                           file_name="Entropy Table"):
    """

        Calculate the entropy of each non-continous numerical feature in a pandas
        dataframe object and store in a pandas dataframe object in the proper
        directory structure.

    Args:
        df: pd.Dataframe
            Pandas DataFrame object

        df_features: DataFrameTypes from eflow
            DataFrameTypes object

        output_folder_path: str
            Pre defined path to already existing directory to output file(s).

        sub_dir: str
            Path to be possibly generated.

        file_name: str
            Name of the given file to save

    Returns:
        Nothing
    """
    entropy_dict = dict()
    for feature_name in df.columns:
        if feature_name in df_features.all_features() and \
                feature_name not in df_features.null_only_features() and \
                feature_name not in df_features.continuous_numerical_features():
            entropy_dict[feature_name] = calculate_entropy(
                df[feature_name].dropna())

    entropy_table = pd.DataFrame.from_dict(entropy_dict,
                                           orient='index').rename(columns={0: "Entropy"})

    entropy_table.index.name = "Features"

    entropy_table.sort_values(by=["Entropy"],
                              ascending=True,
                              inplace=True)

    create_dir_structure(output_folder_path,
                         sub_dir)


    pickle_object_to_file(entropy_table,
                          output_folder_path + sub_dir,
                          file_name)

    df_to_image(entropy_table,
                output_folder_path,
                sub_dir,
                "Entropy Table",
                show_index=True,
                format_float_pos=5)


def auto_binning(df,
                 df_features,
                 feature_name,
                 bins=5):
    """

        Takes a pandas series object and assigns generalized labels and binning
        dimensions.

    Args:
        df: pd.Dataframe
            Pandas Datafrane object

        df_features: DataFrameTypes from eflow
            DataFrameTypes object

        feature_name: string
            Name of the feature to extract the series from

        bins: int
            Number of bins to create.

    Returns:
        Gives back the bins and associated labels
    """

    if feature_name not in df_features.all_features():
        raise KeyError("Feature name must be encapsulated in df_features.")

    if feature_name not in df_features.float_features() and feature_name not in df_features.integer_features():
        raise ValueError("Feature must be a float or an integer to properly bin the given data.")

    # Create bins of type pandas.Interval
    binned_list = list(pd.cut(df[feature_name].dropna().sort_values(),
                              bins).unique())

    # Iterate through all possible bins
    bins = []
    labels = []
    for bin_count, binned_obj in enumerate(binned_list):

        # Extract from pandas.Interval into a list; just nicer to read
        binned_obj = [binned_obj.left, binned_obj.right]

        # Convert to int if the feature is an int type
        if feature_name in df_features.integer_features():
            binned_obj[0] = int(binned_obj[0])
            binned_obj[1] = int(binned_obj[1])

        # -----
        if bin_count == 0:

            # Move bined value down so it properly captures the starting integer
            if feature_name in df_features.integer_features():
                bins.append(int(binned_obj[0]) - .000001)
            else:
                bins.append(binned_obj[0])

        # -----
        if feature_name in df_features.integer_features():

            # Values are the same change label look
            if bin_count != 0 and binned_obj[0] + 1 == binned_obj[1]:
                labels.append("=" + str(binned_obj[1]))
            else:

                if bin_count == 0:
                    labels.append(
                        str(binned_obj[0]) + u" \u27f7 " + str(binned_obj[1]))
                else:
                    labels.append(str(binned_obj[0] + 1) + u" \u27f7 " + str(
                        binned_obj[1]))

            # Move bined value up so it properly captures the ending integer
            binned_obj[1] = int(binned_obj[1]) + .000001
        else:
            labels.append(
                str(binned_obj[0]) + "+ " u"\u27f7 " + str(binned_obj[1]))

        bins.append(binned_obj[1])

    bins = [float(bins[i]) for i in range(0, len(bins))]
    return bins, labels



def value_counts_table(df,
                       feature_name):
    """

        Creates a value counts dataframe.

    Args:
        df:
            Pandas DataFrame object.

        feature_name:
            Specified feature column name.

    Returns:
        Returns back a pandas Dataframe object of a feature's value counts
        with percentages.
    """

    # Value counts DataFrame
    value_count_df = df[feature_name].value_counts().rename_axis(
        'Unique Values').reset_index(name='Counts')

    total_count = sum(df[feature_name].dropna().value_counts().values)
    value_count_df["Percantage"] = ["{0:.4f}%".format(count/total_count * 100)
                                    for value, count in
                                    df[feature_name].value_counts().items()]

    value_count_df.set_index('Unique Values',
                             inplace=True)

    return value_count_df


def descr_table(df,
                feature_name,
                to_numeric=False):
    """

        Creates numerical description of a feature of a dataframe.

    Args:
        df: pd.Dataframe
            Pandas DataFrame object.

        feature_name: string
            Specified feature column name.

        to_numeric: bool
            Converts the pandas series to all numeric.

    Returns:
        Returns back a Dataframe object of a numerical feature's summary.
    """

    # Convert to numeric without raising errors
    if to_numeric:
        errors = "coerce"
    else:
        errors = "ignore"

    desc_df = pd.to_numeric(df[feature_name],
                            errors=errors).dropna().describe().to_frame()
    desc_df.loc["var"] = pd.to_numeric(df[feature_name],
                                       errors=errors).dropna().var()

    return desc_df



def suggest_removal_features(df):
    """

        Will find features that appear to be almost index like with their
        feature values.
    Args:
        df: pd.Dataframe
            Pandas DataFrame object.

    Returns:
        Returns back a list of features to remove.
    """
    features_to_remove = set()

    # Return back index like features
    for feature in df.columns:
        if len(df[feature].value_counts().index.tolist()) >= int(
                df.shape[0] / 2):
            features_to_remove.add(feature)

    return features_to_remove


# from ._hidden.general_objects import enum
# from ._hidden.constants import SYS_CONSTANTS
# from ._hidden.custom_exceptions import UnsatisfiedRequirments
# from ._hidden.parent_objects import FileOutput



from eflow._hidden.parent_objects import FileOutput

class PipelineJupyterWidget(FileOutput):
    def __init__(self,
                 widget_child_name):
        FileOutput.__init__(self,
                            f'_Extras/Pipeline Structure/Widgets/{widget_child_name}')


__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


def enum(**enums):
    """
        Allows for constant like variables.
    """
    return type('Enum', (), enums)


import setuptools
import os
with open("README.md", "r") as fh:
    long_description = fh.read()

print(os.getcwd())

MAIN_PACKAGES = [i.strip() for i in open("extra_setup/main_packages.txt").readlines()]

setuptools.setup(
     name='eflow',
     version='0.2.03',
     scripts=['init_env'],
     author="Eric Cacciavillani",
     author_email="eric.cacciavillani@gmail.com",
     description="",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/EricCacciavillani/eFlow",
     packages=setuptools.find_packages(),
     install_requires=MAIN_PACKAGES,
     classifiers=[
         "Programming Language :: Python :: 3.7",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )


import copy

def create_decorrelate_df(df,
                          df_features,
                          target_name,
                          desired_col_average=0.5,
                          show=True):
    df = copy.deepcopy(df)
    df_features = copy.deepcopy(df_features)
    while True:

        # Display correlation map
        corr_metrics = df.corr()
        if show and PROJECT.IS_NOTEBOOK:
            display(corr_metrics.style.background_gradient())

        # Get the correlation means of each feature
        corr_feature_means = []
        for feature_name in list(corr_metrics.columns):

            # Ignore target feature; Only a problem if target was numerical
            if target_name != feature_name:
                corr_feature_means.append(corr_metrics[feature_name].mean())

        if show:
            # Display graph rank
            display_rank_graph(feature_names=list(corr_metrics.columns),
                               metric=corr_feature_means,
                               title="Average Feature Correlation",
                               y_title="Correlation Average",
                               x_title="Features")

        index, max_val = get_max_index_val(corr_feature_means)

        if max_val > desired_col_average:
            # Drop col and notify
            feature_name = list(corr_metrics.columns)[index]
            df.drop(feature_name, axis=1, inplace=True)
            df_features.remove(feature_name)
            print("Dropped column: {0}".format(feature_name))
            vertical_spacing(5)

        # End loop desired average reached
        else:
            if show and PROJECT.IS_NOTEBOOK:
                display(corr_feature_means)
            break

    return df, df_features


def random_partition_of_random_samples(list_of_df_indexes,
                                       random_sampled_rows,
                                       random_sample_amount,
                                       random_state=None):
    # Convert to numpy array if list
    if isinstance(list_of_df_indexes, list):
        list_of_df_indexes = np.array(list_of_df_indexes)

    np.random.seed(random_state)
    for _ in range(np.random.randint(1, 3)):
        np.random.shuffle(list_of_df_indexes)

    if random_sample_amount > len(list_of_df_indexes):
        random_sample_amount = len(list_of_df_indexes)

    return_matrix = np.zeros((random_sample_amount, random_sampled_rows))
    for i in range(random_sampled_rows):
        sub_list = list_of_df_indexes[:random_sample_amount]
        return_matrix[i] = sub_list
        np.random.shuffle(list_of_df_indexes)

    np.random.seed(None)
    return return_matrix


def kmeans_impurity_sample_removal(df,
                                   target,
                                   pca_perc,
                                   majority_class,
                                   majority_class_threshold=.5,
                                   random_state=None):
    """
        Generate models based on the found 'elbow' of the interia values.
    """
    df = copy.deepcopy(df)
    removal_df_indexes = []

    # Create scaler object
    scaler = StandardScaler()
    scaled = scaler.fit_transform(df.drop(columns=[target]))

    print("\nInspecting scaled results!")
    # self.__inspect_feature_matrix(matrix=scaled,
    #                               feature_names=df.columns)

    pca = PCA()
    scaled = pca.fit_transform(scaled)

    # Generate "dummy" feature names
    pca_feature_names = ["PCA_Feature_" +
                         str(i) for i in range(1,
                                               len(df.columns) + 1)]

    print("\nInspecting applied pca results!")
    # self.__inspect_feature_matrix(matrix=scaled,
    #                               feature_names=pca_feature_names)

    if pca_perc < 1.0:
        # Find cut off point on cumulative sum
        cutoff_index = np.where(
            pca.explained_variance_ratio_.cumsum() > pca_perc)[0][0]
    else:
        cutoff_index = scaled.shape[1] - 1

    print(
        "After applying pca with a cutoff percentage {0}%"
        " for the cumulative index. Using features 1 to {1}".format(
            pca_perc, cutoff_index + 1))

    print("Old shape {0}".format(scaled.shape))

    scaled = scaled[:, :cutoff_index + 1]
    pca_feature_names = pca_feature_names[0: cutoff_index + 1]

    print("New shape {0}".format(scaled.shape))

    scaled = scaler.fit_transform(scaled)

    print("\nInspecting re-applied scaled results!")
    # self.__inspect_feature_matrix(matrix=scaled,
    #                               feature_names=pca_feature_names)

    ks = range(1, 15)
    inertias = []
    all_models = []

    for k in tqdm(ks):
        # Create a KMeans instance with k clusters: model
        model = KMeans(n_clusters=k,
                       random_state=random_state).fit(scaled)

        # Append the inertia to the list of inertias
        inertias.append(model.inertia_)
        all_models.append(model)

    a = KneeLocator(inertias, ks, curve='convex', direction='decreasing')
    elbow_index = np.where(inertias == a.knee)
    best_worst_model_index = elbow_index[0][0] + 2
    best_worst_model = None

    if len(all_models) < best_worst_model_index:
        print("OUT OF INDEX ERROR!!!")
    else:
        best_worst_model = all_models[best_worst_model_index]

    print(best_worst_model.labels_)
    df["Cluster_Name"] = model.labels_
    for val in set(df["Cluster_Name"]):
        sub_df = df[df["Cluster_Name"] == val]
        val_counts_dict = sub_df[target].value_counts().to_dict()

        if majority_class in val_counts_dict and val_counts_dict[
            majority_class] / sum(val_counts_dict.values()) <= majority_class_threshold:
            removal_df_indexes += sub_df[
                sub_df[target] == majority_class].index.values.tolist()

    return removal_df_indexes



# Expermental code designed to prove therories. NOT TO BE USED IN PRODUCTION YET!!!!

# Getting Sklearn Models
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

# Visuals Libs
import matplotlib.pyplot as plt
import pylab as pl
import seaborn as sns
from IPython.display import display, HTML
import imageio
from IPython.display import Image

# Pandas lib
import pandas as pd

# Data Science Libs
from scipy.stats import zscore

# Math Libs
from scipy.spatial import distance
import numpy as np

# System Libs
import os, sys
import shutil
import math
import six
import time
import datetime
from functools import partial
import threading

# Misc Libs
import shelve
import copy
from collections import OrderedDict
from multiprocessing import Pool as ThreadPool
import multiprocessing as mp
from tqdm import tqdm

from eflow._hidden.Functionality.Multi_Threading_Functions import *


class TargetSampleRemoval:
    def __init__(self,
                 df,
                 sample_target_dict,
                 columns_to_drop,
                 apply_pca=True,
                 pca_perc=.8,
                 project_name="Default",
                 overwrite_figure_path=None,
                 show_visuals=True,
                 ):
        """
        df:
            Must be a pandas dataframe object

        sample_target_dict:
            Column name(s) to value(s) in the dataframe to create a pandas
            dataframe with just those value(s).

        columns_to_drop:
            Column names to drop from the dataframe

        apply_pca:
            Data had already

        pca_perc:
            PCA cutoff point

        project_name:
            Starting folder name where the system

        overwrite_figure_path:
            Overwrites the absolute path for the images to be generated
        """

        def enum(**enums):
            return type('Enum', (), enums)

        if overwrite_figure_path:
            output_fig_sub_dir = overwrite_figure_path
        else:
            if pca_perc > 1:
                pca_perc = 1
            output_fig_sub_dir = "/Figures/" + project_name + \
                                 "/SampleRemoval_PCA_Features={0}".format(
                                     pca_perc)

        # Project directory structure
        self.__PROJECT = enum(
            PATH_TO_OUTPUT_FOLDER=''.join(
                os.getcwd().partition('/eflow')[0:1]) + output_fig_sub_dir)

        # Copy dataframes for later use
        df = copy.deepcopy(df)

        # Create dataframe of only target values
        for col, df_value in sample_target_dict.items():

            if isinstance(df_value, int):
                df_value = [df_value]

            for val in df_value:
                df = df[df[col] == val]

        for col in columns_to_drop:
            df.drop(columns=[col],
                    inplace=True)

        # --- Apply pca ---
        if apply_pca:

            # Create scaler object
            scaler = StandardScaler()
            scaled = scaler.fit_transform(df)

            print("\nInspecting scaled results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=df.columns)

            pca, scaled = self.__visualize_pca_variance(scaled, show_visuals)

            # Generate "dummy" feature names
            pca_feature_names = ["PCA_Feature_" +
                                 str(i) for i in range(1,
                                                       scaled.shape[1] + 1)]

            print("\nInspecting applied pca results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=pca_feature_names)

            # Use only some of the features based on the PCA percentage
            if pca_perc < 1.0:
                cutoff_index = np.where(
                    pca.explained_variance_ratio_.cumsum() > pca_perc)[0][0]
            # Use all features
            else:
                cutoff_index = scaled.shape[1] - 1

            print(
                "After applying pca with a cutoff percentage of {0}%"
                " for the cumulative index. Using features 1 to {1}".format(
                    pca_perc, cutoff_index + 1))

            print("Old shape {0}".format(scaled.shape))

            scaled = scaled[:, :cutoff_index + 1]
            pca_feature_names = pca_feature_names[0: cutoff_index + 1]

            print("New shape {0}".format(scaled.shape))

            scaled = scaler.fit_transform(scaled)

            print("\nInspecting re-applied scaled results!")
            self.__inspect_feature_matrix(matrix=scaled,
                                          feature_names=pca_feature_names)

            self.__org_scaled = copy.deepcopy(scaled)
            self.__scaled = copy.deepcopy(scaled)

            self.__feature_weights = np.array(pca.explained_variance_ratio_[
                                              :scaled.shape[1]])

            print(self.__feature_weights)


            self.__feature_degress = (self.__feature_weights/self.__feature_weights.sum()) * 9

            print(self.__feature_degress)


        # Assumed PCA has already been applied; pass as matrix
        else:
            self.__scaled = df.values

        new_folder_path = ''.join(
            os.getcwd().partition('/eflow')[0:1]) + "/Figures/" + \
                          project_name + "/SampleRemoval_PCA_Features={0}".format(
            scaled.shape[1])

        if not os.path.exists(new_folder_path):
            os.rename(self.__PROJECT.PATH_TO_OUTPUT_FOLDER,
                      new_folder_path)
        else:
            shutil.rmtree(self.__PROJECT.PATH_TO_OUTPUT_FOLDER)
        self.__PROJECT = enum(
            PATH_TO_OUTPUT_FOLDER=new_folder_path)
        self.__targeted_df = copy.deepcopy(df)

        # Init dummy variables to only be used for multithreading
        self.__index_array = None
        self.__total_indexes = None
        self.__tmp_reduced_scaled = None
        self.__all_dp_dist_list = None
        self.__removed_dps_dict = dict()
        self.__org_df_index_dict = None
        self.__saved_pic_paths_dict = dict()
        self.__applied_methods = set()
        self.__pbar = None

    def __weighted_eudis(self,
                         v1,
                         v2):
        dist = [((a - b) ** 2) * w for a, b, w in zip(v1, v2,
                                                      self.__feature_weights)]
        dist = math.sqrt(sum(dist))
        return dist

    def __rotate_point(self,
                       origin,
                       point,
                       angle):
        """
        Rotate a point counterclockwise by a given angle around a given origin.

        The angle should be given in radians.

        # Author link: http://tinyurl.com/y4yz5hco
        """
        ox, oy = origin
        px, py = point

        qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
        qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
        return qx, qy

    # Not created by me!
    # Created by my teacher: Narine Hall
    # MODIFIED
    def __visualize_pca_variance(self,
                                 data,
                                 show_visuals):
        """
            Visualize PCA matrix feature importance
        """

        # Check for pca variance
        pca = PCA()
        data = pca.fit_transform(data)

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_)
        plt.xticks()
        plt.ylabel('variance ratio')
        plt.xlabel('PCA feature')
        plt.tight_layout()
        self.__create_plt_png("PCA", "PCA_Feature_Variance_Ratio")
        if show_visuals:
            plt.show()
        plt.close()

        # ----
        plt.bar(range(1, len(pca.explained_variance_ratio_) + 1),
                pca.explained_variance_ratio_.cumsum())
        plt.xticks()
        plt.ylabel('cumulative sum of variances')
        plt.xlabel('PCA feature')
        plt.tight_layout()
        self.__create_plt_png("PCA", "PCA_Cumulative_Sum_of_Variances")
        if show_visuals:
            plt.show()
        plt.close()

        return pca, data

    def __check_create_figure_dir(self,
                                  sub_dir):
        """
            Checks/Creates required directory structures inside
            the parent directory figures.
        """

        directory_pth = self.__PROJECT.PATH_TO_OUTPUT_FOLDER

        for dir in sub_dir.split("/"):
            directory_pth += "/" + dir
            if not os.path.exists(directory_pth):
                os.makedirs(directory_pth)

        return directory_pth

    def __create_plt_png(self,
                         sub_dir,
                         filename):
        """
            Saves the plt based image in the correct directory.
        """
        # Ensure directory structure is init correctly
        abs_path = self.__check_create_figure_dir(sub_dir)

        # Ensure file ext is on the file.
        if filename[-4:] != ".png":
            filename += ".png"

        fig = plt.figure(1)
        fig.savefig(abs_path + "/" + filename, bbox_inches='tight')

    def __inspect_feature_matrix(self,
                                 matrix,
                                 feature_names):
        """
            Creates a dataframe to quickly analyze a matrix
        """
        mean_matrix = np.mean(matrix, axis=0)
        std_matrix = np.std(matrix, axis=0)
        data_dict = dict()
        for index, feature_name in enumerate(feature_names):
            data_dict[feature_name] = [mean_matrix[index],
                                       std_matrix[index]]

        tmp_df = pd.DataFrame.from_dict(data_dict,
                                        orient='index',
                                        columns=['Mean', 'Standard Dev'])

        display(tmp_df)

        return tmp_df

    def remove_noise(self,
                     new_sample_amount,
                     zscore_high=2.0,
                     annotate=False,
                     apply_changes=False,
                     display_all_graphs=False,
                     show_gif=False,
                     shelve_relative_path=None,
                     create_visuals=True):

        new_sample_amount = int(new_sample_amount)

        if new_sample_amount >= self.__scaled.shape[0]:
            print("THROW ERROR HERE: Sample removal must be less then")
            return
        elif new_sample_amount <= 0:
            print("THROW ERROR HERE: Val must be a positive number!")
            return
        else:

            # Display graph before augmentation; Create centroid
            centroid = np.mean(self.__scaled, axis=0)
            column_list = [i for i in range(0, self.__scaled.shape[1])]

            df_index_scaled_dict = dict()

            reduced_scaled = np.column_stack(
                (self.__scaled, self.__targeted_df.index.values.reshape(
                    (self.__scaled.shape[0], 1)).astype(self.__scaled.dtype)))

            # Index to shape
            if not self.__org_df_index_dict:
                self.__org_df_index_dict = dict()
                for i, df_index in enumerate(self.__targeted_df.index.values):
                    df_index = reduced_scaled[i][-1]
                    self.__org_df_index_dict[df_index] = i

            for i, _ in enumerate(reduced_scaled):
                df_index = reduced_scaled[i][-1]
                df_index_scaled_dict[df_index] = i

            if create_visuals:

                if zscore_high:
                    folder_dir_name = "Data_Point_Removal_Noise_Zscore={0}".format(
                        zscore_high)
                else:
                    folder_dir_name = "Data_Point_Removal_Noise_Zscore=NaN"

                if apply_changes:
                    self.__applied_methods.add(folder_dir_name)

                self.__visualize_data_points(centroid=centroid,
                                             scaled_data=self.__scaled,
                                             new_sample_amount=new_sample_amount,
                                             annotate=annotate,
                                             apply_changes=apply_changes,
                                             output_path=folder_dir_name,
                                             called_from=sys._getframe().f_code.co_name,
                                             title="Starting point",
                                             display_all_graphs=display_all_graphs)

            dp_distances = np.zeros(len(reduced_scaled))

            if "Remove Noise" not in self.__removed_dps_dict.keys():
                self.__removed_dps_dict["Remove Noise"] = list()

        # Keep looping until new sample amount has been reached or
        # the distances are properly.
        while reduced_scaled.shape[0] > new_sample_amount:

            for index, dp in enumerate(reduced_scaled):
                dp_distances[index] = self.__weighted_eudis(
                    centroid, dp[:column_list[-1] + 1])

            farthest_dp_index = np.argmax(dp_distances)
            zscores_dp_distances = zscore(np.concatenate((
                dp_distances, np.array([self.__weighted_eudis(
                    centroid,self.__org_scaled[self.__org_df_index_dict[
                        dp_index]])
                                        for dp_index in
                                        self.__removed_dps_dict["Remove Noise"]
                                        ])), axis=0))

            if zscores_dp_distances[farthest_dp_index] >= zscore_high:

                # Add original dataframe index to the dict;
                # remove actual row from the data
                df_index = int(reduced_scaled[farthest_dp_index][-1])
                self.__removed_dps_dict["Remove Noise"].append(df_index)

                if shelve_relative_path:
                    shelf = shelve.open(shelve_relative_path)
                    shelf[shelve_relative_path.split("/")[-1]] = list(
                        self.__removed_dps_dict["Remove Noise"])
                    shelf.close()

                reduced_scaled = np.delete(reduced_scaled,
                                           farthest_dp_index,
                                           0)
                # Update centroid
                centroid = np.mean(reduced_scaled[:, column_list],
                                   axis=0)
                if create_visuals:

                    meta_data = dict()
                    meta_data["zscore"] = zscores_dp_distances[
                        farthest_dp_index]
                    meta_data["distance"] = dp_distances[
                        farthest_dp_index]

                    self.__visualize_data_points(centroid=centroid,
                                                 scaled_data=reduced_scaled[
                                                             :,
                                                             column_list],
                                                 new_sample_amount=new_sample_amount,
                                                 annotate=annotate,
                                                 apply_changes=apply_changes,
                                                 output_path=folder_dir_name,
                                                 meta_data=meta_data,
                                                 called_from=sys._getframe().f_code.co_name,
                                                 title="Data Removal: Noise reduction",
                                                 display_all_graphs=display_all_graphs)
                else:
                    print(
                        "Scaled size is now {0} and Z-Score {1:.2f}.".format(
                            reduced_scaled.shape[0],
                            zscores_dp_distances[farthest_dp_index]))
            # Break loop distances are below z-score val
            else:
                break

        if create_visuals:
            self.__create_gif_with_dp_amount(n_start=self.__scaled.shape[0],
                                             n_end=reduced_scaled.shape[0],
                                             folder_dir_name=folder_dir_name,
                                             filename="Noise Reduction",
                                             show_gif=show_gif)

        df_removal_indexes = copy.deepcopy(
            self.__removed_dps_dict["Remove Noise"])
        if apply_changes:
            self.__scaled = reduced_scaled[:, column_list]

            for i in df_removal_indexes:
                try:
                    self.__targeted_df.drop(i, inplace=True)
                except KeyError:
                    pass
        else:
            self.__removed_dps_dict.pop("Remove Noise", None)

        return df_removal_indexes

    def remove_similar(self,
                       new_sample_amount,
                       weighted_dist_value=1.0,
                       annotate=False,
                       apply_changes=False,
                       display_all_graphs=False,
                       show_gif=False,
                       shelve_relative_path=None,
                       create_visuals=True):

        self.__index_array = None
        self.__total_indexes = None
        self.__tmp_reduced_scaled = None
        self.__all_dp_dist_list = None
        self.__pbar = None
        self.__all_dp_dist_dict = None

        new_sample_amount = int(new_sample_amount)

        if new_sample_amount >= self.__scaled.shape[0]:
            print("THROW ERROR HERE: Sample removal must be less then")
        elif new_sample_amount <= 0:
            print("THROW ERROR HERE: Val must be a positive number!")
        else:

            df_index_scaled_dict = dict()
            reduced_scaled = np.column_stack(
                (self.__scaled, self.__targeted_df.index.values.reshape(
                    (self.__scaled.shape[0], 1)).astype(self.__scaled.dtype)))

            # Index to shape
            if not self.__org_df_index_dict:
                self.__org_df_index_dict = dict()
                for i, df_index in enumerate(self.__targeted_df.index.values):
                    df_index = reduced_scaled[i][-1]
                    self.__org_df_index_dict[df_index] = i

            for i, _ in enumerate(reduced_scaled):
                df_index = reduced_scaled[i][-1]
                df_index_scaled_dict[df_index] = i

            # Display graph before augmentation; Create centroid
            centroid = np.mean(self.__scaled, axis=0)
            column_list = [i for i in range(0, self.__scaled.shape[1])]

            for i, _ in enumerate(reduced_scaled):
                df_index = reduced_scaled[i][-1]
                df_index_scaled_dict[df_index] = i

            if create_visuals:

                if weighted_dist_value:
                    folder_dir_name = "Data_Point_Removal_Similar_Weight={0}".format(
                        weighted_dist_value)
                else:
                    folder_dir_name = "Data_Point_Removal_Similar_Weight=NaN"

                if apply_changes:
                    self.__applied_methods.add(folder_dir_name)
                self.__visualize_data_points(centroid=centroid,
                                             scaled_data=self.__scaled,
                                             new_sample_amount=new_sample_amount,
                                             annotate=annotate,
                                             apply_changes=apply_changes,
                                             output_path=folder_dir_name,
                                             called_from=sys._getframe().f_code.co_name,
                                             title="Starting point",
                                             display_all_graphs=display_all_graphs)

            starting_shape = reduced_scaled.shape[0]

            if "Remove Similar" not in self.__removed_dps_dict.keys():
                self.__removed_dps_dict["Remove Similar"] = list()

            farthest_dp_distance = None
            dp_distances = np.zeros(len(reduced_scaled))

            while reduced_scaled.shape[0] > new_sample_amount:
                # Following unconventional programming for multi threading
                # speed and memory increase
                self.__index_array = [i for i in
                                      range(0, len(reduced_scaled))]
                self.__total_indexes = len(self.__index_array)
                self.__tmp_reduced_scaled = copy.deepcopy(
                    reduced_scaled[:, column_list])

                if not farthest_dp_distance:
                    for index, dp in enumerate(self.__tmp_reduced_scaled):
                        dp_distances[index] = self.__weighted_eudis(
                            centroid, dp[:column_list[-1] + 1])

                    farthest_dp_distance = np.amax(dp_distances)
                    farthest_dp_distance *= weighted_dist_value

                removal_index, keep_index, smallest_dist = self.__shortest_dist_relationship(
                    centroid)

                if farthest_dp_distance < smallest_dist:
                    print("Target distance reached!!!")
                    break

                df_index = int(reduced_scaled[removal_index][-1])
                self.__removed_dps_dict["Remove Similar"].append(df_index)

                if shelve_relative_path:
                    shelf = shelve.open(shelve_relative_path)
                    shelf[shelve_relative_path.split("/")[-1]] =\
                        self.__removed_dps_dict["Remove Similar"]
                    shelf.close()

                # Remove from temp scaled
                reduced_scaled = np.delete(reduced_scaled,
                                           removal_index,
                                           0)
                # Update centroid
                centroid = np.mean(reduced_scaled[:, column_list],
                                   axis=0)

                if create_visuals:

                    meta_data = dict()
                    meta_data["kept_point"] = self.__tmp_reduced_scaled[
                        keep_index]

                    self.__visualize_data_points(centroid=centroid,
                                                 scaled_data=reduced_scaled[
                                                             :,
                                                             column_list],
                                                 new_sample_amount=new_sample_amount,
                                                 annotate=annotate,
                                                 apply_changes=apply_changes,
                                                 output_path=folder_dir_name,
                                                 called_from=sys._getframe().f_code.co_name,
                                                 meta_data=meta_data,
                                                 title="Data Removal: Similarity removal",
                                                 display_all_graphs=display_all_graphs)
                else:
                    print("Scaled size is now {0}.".format(
                        reduced_scaled.shape[0]))

        # De-init multithreading artifacts
        self.__index_array = None
        self.__total_indexes = None
        self.__tmp_reduced_scaled = None
        self.__all_dp_dist_list = None

        if create_visuals:
            self.__create_gif_with_dp_amount(n_start=starting_shape - 1,
                                             n_end=reduced_scaled.shape[0],
                                             folder_dir_name=folder_dir_name,
                                             filename="Similar Reduction",
                                             show_gif=show_gif)

        df_removal_indexes = copy.deepcopy(self.__removed_dps_dict["Remove Similar"])
        if apply_changes:
            self.__scaled = reduced_scaled[:, column_list]
            for i in df_removal_indexes:
                try:
                    self.__targeted_df.drop(i, inplace=True)
                except KeyError:
                    pass
        else:
            self.__removed_dps_dict.pop("Remove Similar", None)

        return df_removal_indexes

    def __find_dp_dist_mean(self,
                            target_index,
                            index_array,
                            scaled_data):
        distances = np.zeros(len(index_array))
        for index, dp_index in enumerate(
                filter(lambda x: x != target_index, index_array)):
            distances[index] = self.__weighted_eudis(
                scaled_data[target_index],
                scaled_data[dp_index])

        return np.mean(distances)

    def __shortest_dist_with_target(self,
                                    target_dp_index):
        """
            Finds the shortest distance between all dps based on its postional.
        """

        distances = np.zeros(self.__total_indexes - (target_dp_index + 1))
        for index, dp_index in enumerate(self.__index_array[
                                         target_dp_index + 1:]):
            distances[index] = self.__weighted_eudis(self.__tmp_reduced_scaled[
                                                         target_dp_index],
                                                     self.__tmp_reduced_scaled[
                                                         dp_index])
        shortest_dp_index = np.argmin(distances)

        return {
            target_dp_index: (
            self.__index_array[target_dp_index + 1:][shortest_dp_index],
            distances[shortest_dp_index])}

    def __shortest_dist_relationship(self,
                                     centroid):

        """
            Finds the two datapoints that have the smallest distance.
        """
        if not self.__all_dp_dist_list:

            total = 0
            for i in range(0,
                           self.__tmp_reduced_scaled.shape[0]):
                total += i

            print("The total time required is:", str(
                datetime.timedelta(seconds=total * 1.3e-5)))

            self.__all_dp_dist_list = find_all_distances_in_matrix(
                matrix=self.__tmp_reduced_scaled,
                index_array=self.__index_array,
                total_indexes=self.__total_indexes,
                feature_weights=self.__feature_weights)

        # :::ADD WEIGHTED DISTANCE IDEA HERE FUTURE ERIC:::

        all_shortest = [
            [target_dp_index,
             np.argmin(distances) + target_dp_index + 1,
             np.amin(distances)]
            for target_dp_index, distances in
            enumerate(self.__all_dp_dist_list)
            if len(distances) > 0]

        smallest_dps_relationship = min(all_shortest, key=lambda x: x[2])

        dp_1_index = smallest_dps_relationship[0]
        dp_2_index = smallest_dps_relationship[1]
        smallest_distance = smallest_dps_relationship[2]

        dp_1_dist = self.__weighted_eudis(self.__tmp_reduced_scaled[
                                              dp_1_index],
                                          centroid)

        dp_2_dist = self.__weighted_eudis(self.__tmp_reduced_scaled[
                                              dp_2_index],
                                          centroid)

        # Decide of the two dps which to remove
        removal_index = None
        keep_index = None
        if dp_1_dist < dp_2_dist:
            removal_index = dp_2_index
            keep_index = dp_1_index
        else:
            removal_index = dp_1_index
            keep_index = dp_2_index

        # Return distances values to everyone above the removed index
        for sub_removal_index, dp_index_key in enumerate(
                range(removal_index - 1, -1, -1)):
            self.__all_dp_dist_list[dp_index_key] = np.delete(
                self.__all_dp_dist_list[dp_index_key],
                sub_removal_index, 0)

        self.__all_dp_dist_list.pop(removal_index)

        # Return back the indexes and distance
        return removal_index, keep_index, smallest_distance

    def __create_gif_with_dp_amount(self,
                                    n_start,
                                    n_end,
                                    folder_dir_name,
                                    filename,
                                    flash_final_results=False,
                                    show_gif=False):
        """
            Generates a gif based on pre-generated images of sample removal.
        """

        if folder_dir_name:
            images = [imageio.imread(self.__PROJECT.PATH_TO_OUTPUT_FOLDER + "/" +
                                     folder_dir_name
                                     +
                                     "/Sample_removal_Visualized_Cluster_n={"
                                     "0}.png".format(i)) for i in range(n_start,
                                                                        n_end - 1,
                                                                        -1)]
        else:
            images = [imageio.imread(self.__saved_pic_paths_dict[i])
                                     for i in range(n_start,
                                                    n_end - 1,
                                                    -1)]

        if flash_final_results:

            if folder_dir_name:
                images += [imageio.imread(
                    self.__PROJECT.PATH_TO_OUTPUT_FOLDER + "/" +
                    folder_dir_name
                    + "/Sample_removal_Visualized_Cluster_n={0}.png".format(
                        n_start)), imageio.imread(
                    self.__PROJECT.PATH_TO_OUTPUT_FOLDER + "/" +
                    folder_dir_name
                    + "/Sample_removal_Visualized_Cluster_n={0}_White_Outed.png".format(
                        n_end))] * 4

            else:
                images += [imageio.imread(
                    self.__saved_pic_paths_dict[n_start]),
                              imageio.imread(
                                  self.__saved_pic_paths_dict[n_end])] * 4

        if folder_dir_name:
            imageio.mimsave(self.__PROJECT.PATH_TO_OUTPUT_FOLDER + "/" +
                            folder_dir_name +
                            "/{0}.gif".format(filename),
                            images,
                            duration=.68)
        else:
            imageio.mimsave(self.__PROJECT.PATH_TO_OUTPUT_FOLDER +
                            "/{0}.gif".format(filename),
                            images,
                            duration=.68)

        if show_gif:

            if folder_dir_name:
                display(Image(filename=self.__PROJECT.PATH_TO_OUTPUT_FOLDER +
                                       "/" + folder_dir_name +
                                       "/{0}.gif".format(filename)))
            else:
                display(Image(filename=self.__PROJECT.PATH_TO_OUTPUT_FOLDER +
                                       "/{0}.gif".format(filename)))

    def create_gif_with_dp_amount(self,
                                  n_start,
                                  n_end,
                                  filename=None,
                                  flash_final_results=False,
                                  show_gif=False):

        if not filename:
            filename = ""
            for given_method in self.__applied_methods:
                filename += given_method + " "

        self.__create_gif_with_dp_amount(n_start,
                                         n_end,
                                         folder_dir_name=None,
                                         filename=filename,
                                         flash_final_results=flash_final_results,
                                         show_gif=show_gif)

    def __visualize_data_points(self,
                                centroid,
                                scaled_data,
                                new_sample_amount,
                                annotate,
                                output_path,
                                called_from,
                                title,
                                apply_changes,
                                meta_data=None,
                                white_out_mode=False,
                                display_all_graphs=False,
                                no_print_output=False):
        """
            Creates visualization of clustering model on given data.
                Ex: The parameters are fairly self explanatory.
        """
        pl.figure(figsize=(8, 7))
        pl.title(title + "_n={0}".format(len(scaled_data),
                                         fontsize=15))

        plt.gcf().text(.94, .94, "Target_n_samples={0}".format(
            new_sample_amount), fontsize=12)

        # legennd_string = ""
        # for count, given_method in enumerate(self.__applied_methods):
        #     legennd_string += given_method.split('_', -1)[-1] + " "
        #
        #     if count % 2 == 0:
        #         legennd_string += "\n"
        #
        # plt.gcf().text(.91, 0.9,
        #                legennd_string,
        #                fontsize=12)

        cell_information = np.array([])
        row_information = np.array([])

        for method_count, given_method in enumerate(self.__applied_methods):
            cell_information = np.append(cell_information,
                                         given_method)
            row_information = np.append(row_information,
                                        "Removal Process {0}".format(
                                            method_count))
        cell_information = cell_information.reshape(len(
            self.__applied_methods), 1)

        # plt.axis("off")

        # the_table = plt.table(cellText=cell_information,
        #                       rowLabels=row_information,
        #                       colLabels=np.array(["Table"]),
        #                       colWidths=[0.5] * 3,
        #                       loc='center left',
        #                       bbox=[1.3, -0.5, 0.5, 0.5],
        #                       fontsize=14)
        plt.subplots_adjust(bottom=0.3)
        # plt.show()

        # Plot existing data points
        for i in range(0, scaled_data.shape[0]):

            rotation_degrees = (
                    (abs(centroid - scaled_data[i])/2) * self.__feature_degress).sum()

            px, py = self.__rotate_point(np.array([0,0]),
                                         np.array(
                                             [0,
                                              self.__weighted_eudis(
                                                  scaled_data[i], centroid)]),
                                         rotation_degrees)

            pl.scatter(px,
                       py,
                       c="#0080FF",
                       marker='o',
                       label="Existing data point")

            # plt.annotate('{0}'.format(i), xy=(np.mean(scaled_data[i]),
            #            distance.euclidean(scaled_data[i], centroid)),
            #              xytext=(np.mean(scaled_data[i]),
            #                      distance.euclidean(scaled_data[i], centroid)))

        # Plot data points removed from noise removal
        for key_name,list_of_removed_indexes in self.__removed_dps_dict.items():

            last_index = len(list_of_removed_indexes) - 1
            for index, dp_index in enumerate(list_of_removed_indexes):
                if white_out_mode:

                    rotation_degrees = (
                            (abs(centroid - dp)/2) * self.__feature_degress).sum()

                    px, py = self.__rotate_point(np.array([0,0]),
                                                 np.array(
                                                     [0,
                                                      self.__weighted_eudis(
                                                          self.__org_scaled[
                                                              dp_index])]),
                                                 rotation_degrees)

                    pl.scatter(px,
                               py,
                               c="#ffffff",
                               marker='X',
                               alpha=0)
                else:

                    if key_name == "Remove Noise":

                        dp = self.__org_scaled[self.__org_df_index_dict[
                            dp_index]]

                        rotation_degrees = (
                                (abs(centroid - dp)/2) * self.__feature_degress).sum()

                        px, py = self.__rotate_point(np.array([0,0]),
                                                     np.array([0,
                                                               self.__weighted_eudis(
                                                                   dp,
                                                                   centroid)]),
                                                     rotation_degrees)

                        pl.scatter(px,
                                   py,
                                   c="#00A572",
                                   marker='X',
                                   label="Noise Removal")

                        if annotate and meta_data \
                                and index == last_index \
                                and called_from == "remove_noise":

                            dp = self.__org_scaled[
                                self.__org_df_index_dict[dp_index]]

                            # Find the correct angle to have the text and annotated line match
                            dp_centroid_dist = self.__weighted_eudis(dp,
                                                                     centroid)

                            dy = (0 - py)
                            dx = (0 - px)
                            rotn = np.degrees(np.arctan2(dy, dx))
                            trans_angle = plt.gca().transData.transform_angles(
                                np.array((rotn,)), np.array((dx,
                                                             dy)).reshape(
                                    (1, 2)))[0]
                            # Fix text representation on the given angle
                            if trans_angle > 90:
                                trans_angle = trans_angle - 180

                            if trans_angle < -90:
                                trans_angle = trans_angle + 180

                            # Spacing for formatting
                            spacing = "\n" * 2
                            if trans_angle < 0:
                                spacing = "\n" * 4

                            # Create line
                            plt.annotate(' ', xy=(px,
                                                  py),
                                         xytext=(0,
                                                 0),
                                         rotation=trans_angle,
                                         ha='center',
                                         va='center',
                                         rotation_mode='anchor',
                                         arrowprops={'arrowstyle': '<->',
                                                     'shrinkA': .4,
                                                     'shrinkB': 4.5}
                                         )

                            # Create text
                            plt.annotate(
                                spacing + 'zscore={0:.2f} , Dist:{1:.2f}\n'.format(
                                    meta_data["zscore"],
                                    meta_data["distance"]),
                                xy=(0, 0),
                                rotation_mode='anchor',
                                va='center',
                                ha='center',
                                rotation=trans_angle)

                    elif key_name == "Remove Similar":
                        dp = self.__org_scaled[self.__org_df_index_dict[
                            dp_index]]

                        rotation_degrees = (
                                (abs(centroid - dp)/2) * self.__feature_degress).sum()

                        px, py = self.__rotate_point(np.array([0, 0]),
                                                     np.array([0,
                                                               self.__weighted_eudis(
                                                                   dp,
                                                                   centroid)]),
                                                     rotation_degrees)

                        pl.scatter(px,
                                   py,
                                   c="#8A2BE2",
                                   marker='X',
                                   label="Similar Removal")

                        if annotate and meta_data \
                                and index == last_index \
                                and called_from == "remove_similar":
                            rotation_degrees = (
                                    (abs(centroid - meta_data["kept_point"])/2) *
                                    self.__feature_degress).sum()
                            meta_px, meta_py = self.__rotate_point(np.array([0, 0]),
                                                                   np.array([0,
                                                                             self.__weighted_eudis(
                                                                                 meta_data[
                                                                                     "kept_point"],
                                                                                 centroid)]),
                                                                   rotation_degrees)
                            # Create line
                            plt.annotate(' ',
                                         xy=(px,
                                             py),
                                         xytext=(
                                             meta_px,
                                             meta_py),
                                         ha='center',
                                         va='center',
                                         rotation_mode='anchor',
                                         arrowprops={'arrowstyle': '<->',
                                                     'shrinkA': .4,
                                                     'shrinkB': 4.5}
                                         )
                            from scipy.spatial import distance
                            print(distance.euclidean((px,py),(meta_px, meta_py)))

        # Plot centroid
        pl.scatter(0, 0,
                   c="r", marker="D",
                   label="Centroid")
        # Author: http://tinyurl.com/yxvd33t2
        # Removes all duplicated handles and labels of the labels
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = OrderedDict(zip(labels, handles))
        plt.legend(by_label.values(), by_label.keys(),
                   loc='center left', bbox_to_anchor=(1, 0.92))

        filename_format = None
        if white_out_mode:
            filename_format = "Sample_removal_Visualized_Cluster_n={0}_White_Outed".format(
                                      len(scaled_data))

        else:
            filename_format = "Sample_removal_Visualized_Cluster_n={0}".format(
                                      len(scaled_data))

        self.__create_plt_png(output_path,
                              filename_format)

        if apply_changes:
            self.__saved_pic_paths_dict[
                len(scaled_data)] = self.__PROJECT.PATH_TO_OUTPUT_FOLDER + \
                                    "/" + output_path + "/" + filename_format\
                                    + ".png"
        if display_all_graphs:
            plt.show()
        else:
            # if self.__scaled.shape[0] > scaled_data.shape[0] and \
            #         not no_print_output:
            #     if new_dp_meta_noise_removal:
            #         print("Scaled size is now {0}".format(
            #             scaled_data.shape[
            #                 0]) + " and Z-Score of {0:.2f}".format(
            #             new_dp_meta_noise_removal[1]))

            # print("Scaled size is now {0}".format(scaled_data.shape[0]
            #                                               ))
            pass

        plt.close()

    # --- Misc
    # # I am this lazy yes...
    # def __vertical_spacing(self, spaces=1):
    #     for _ in range(0, spaces):
    #         print()

    # --- Getters/Setters
    def get_scaled_data(self):
        return copy.deepcopy(self.__scaled)

    def testing_table_data(self):
        return copy.deepcopy(self.__applied_methods)






from eflow.utils.sys_utils import *
from eflow.utils.pandas_utils import df_to_image
from eflow.utils.image_processing_utils import create_plt_png
from eflow._hidden.parent_objects import ModelAnalysis
from eflow._hidden.custom_exceptions import RequiresPredictionMethods, ProbasNotPossible, \
    UnsatisfiedRequirments
from eflow.data_analysis import FeatureAnalysis
from eflow.data_pipeline_segments import DataEncoder

from eflow._hidden.constants import GRAPH_DEFAULTS

from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import recall_score
from sklearn.metrics import precision_score
from sklearn.metrics import classification_report
import scikitplot as skplt
import numpy as np
import warnings
import copy
import pandas as pd
from IPython.display import display
import matplotlib.pyplot as plt
import seaborn as sns

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"

class ClassificationAnalysis(ModelAnalysis):

    """
        Analyzes a classification model's result's based on the prediction
        function(s) passed to it. Creates graphs and tables to be saved in directory
        structure.
    """

    def __init__(self,
                 dataset_name,
                 model,
                 model_name,
                 feature_order,
                 target_feature,
                 pred_funcs_dict,
                 df_features,
                 sample_data,
                 project_sub_dir="Classification Analysis",
                 overwrite_full_path=None,
                 target_classes=None,
                 save_model=True,
                 notebook_mode=False):
        """
        Args:
            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            model:
                A fitted supervised machine learning model.

            model_name:
                The name of the model in string form.

            feature_order: collection object
                Features names in proper order to re-create the pandas dataframe.

            pred_funcs_dict:
                A dict of the name of the function and the function defintion for the
                model prediction methods.
                (Can handle either a return of probabilities or a singile value.)
                Init Example:
                pred_funcs = dict()
                pred_funcs["Predictions"] = model.predict
                pred_funcs["Probabilities"] = model.probas

            sample_data:
                Given data to then pass into our prediction functions to get a
                resultant to get the classification prediction 'type'.
                Can be a matrix or a vector.

            project_sub_dir:
                Creates a parent or "project" folder in which all sub-directories
                will be inner nested.

            overwrite_full_path:
                Overwrites the path to the parent folder.

            target_classes:
                Specfied list/np.array of targeted classes the model predicts. If set to
                none then it will attempt to pull from the sklearn's default attribute
                '.classes_'.

            df_features:
                DataFrameTypes object; organizes feature types into groups.
        """

        # Init parent object
        ModelAnalysis.__init__(self,
                               f'{dataset_name}/{project_sub_dir}/Target Feature: {target_feature}/{model_name}',
                               overwrite_full_path)

        # Init objects without pass by refrence

        # Remove target feature from feature order when trying to recreate dataframe
        self.__target_feature = copy.deepcopy(target_feature)
        self.__feature_order = copy.deepcopy(feature_order)

        if self.__target_feature in self.__feature_order:
            self.__feature_order.remove(self.__target_feature)

        del feature_order
        del target_feature

        self.__model = copy.deepcopy(model)

        self.__model_name = copy.deepcopy(model_name)
        self.__target_values = copy.deepcopy(target_classes)
        self.__df_features = copy.deepcopy(df_features)
        self.__pred_funcs_dict = copy.deepcopy(pred_funcs_dict)
        self.__pred_funcs_types = dict()
        self.__notebook_mode = copy.deepcopy(notebook_mode)

        # Determines if the perform was called
        self.__called_from_perform = False

        # Init on sklearns default target classes attribute
        if not self.__target_values:
            self.__target_values = copy.deepcopy(model.classes_)
        # ---
        if len(self.__target_values) != 2:
            self.__binary_classifcation = False
        else:
            self.__binary_classifcation = True

        # Attempt to save machine learning model
        try:
            if save_model:
                pickle_object_to_file(self.__model,
                                      self.folder_path,
                                      f'{self.__model_name}')
        except:
            pass
        # ---
        create_dir_structure(self.folder_path,
                             "_Extras")
        # Save predicted classes
        write_object_text_to_file(self.__target_values,
                                  self.folder_path + "_Extras",
                                  "_Classes")

        # Save feature order
        write_object_text_to_file(self.__feature_order,
                                  self.folder_path + "_Extras",
                                  "_Feature_Order")


        # Save features and or df_features object
        df_features.create_json_file_representation(self.folder_path + "_Extras",
                                                    "df_features.json")

        self.__sample_data = None

        # Extract sample data
        if len(sample_data.shape) == 2:
            self.__sample_data = np.reshape(sample_data[0],
                                            (-1, sample_data.shape[1]))
        elif len(sample_data.shape) == 1:
            self.__sample_data = [sample_data]
        else:
            raise UnsatisfiedRequirments("This program can only handle 1D and 2D matrices.")


        # Find the 'type' of each prediction. Probabilities or Predictions
        if self.__pred_funcs_dict:
            for pred_name, pred_func in self.__pred_funcs_dict.items():
                try:
                    model_output = pred_func(self.__sample_data)[0]
                except Exception as e:
                    model_output = pred_func(np.array(self.__sample_data))[0]

                # Confidence / Probability (Continuous output)
                if isinstance(model_output, list) or isinstance(model_output,
                                                                np.ndarray):
                    self.__pred_funcs_types[pred_name] = "Probabilities"

                # Classification (Discrete output)
                else:
                    self.__pred_funcs_types[pred_name] = "Predictions"
        else:
            raise RequiresPredictionMethods("This object requires you to pass the prediction methods in a dict with the name of the method as the key.")


        try:
            feature_importances = model.feature_importances_

            self.graph_model_importances(copy.deepcopy(self.__feature_order),
                                         feature_importances,
                                         display_visuals=False)

        except AttributeError:
            pass


    def get_predictions_names(self):
        return self.__pred_funcs_dict.keys()

    def perform_analysis(self,
                         X,
                         y,
                         dataset_name,
                         thresholds_matrix=None,
                         classification_error_analysis=False,
                         classification_correct_analysis=False,
                         ignore_metrics=[],
                         custom_metrics_dict=dict(),
                         average_scoring=["micro",
                                          "macro",
                                          "weighted"],
                         display_visuals=True):
        """

            Runs all available analysis functions on the models predicted data.

        Args:
            X:
                Feature matrix.

            y:
                Target data vector.

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds_matrix:
                List of list/matrix of thresholds.

                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            classification_error_analysis: bool
                Perform feature analysis on data that was incorrectly predicted.

            classification_correct_analysis: bool
                Perform feature analysis on data that was correctly predicted.

            figsize:
                All plot's dimension's.

            ignore_metrics:
                Specify the default metrics to not apply to the classification
                data_analysis.
                    * Precision
                    * MCC
                    * Recall
                    * F1-Score
                    * Accuracy

            custom_metrics_dict:
                Pass the name of metric(s) with the function definition(s) in a
                dictionary.

            average_scoring:
                Determines the type of averaging performed on the data.

            display_visuals:
                Controls visual display of error error data_analysis if it is able to run.

        Returns:
            Performs all classification functionality with the provided feature
            data and target data.
                * plot_precision_recall_curve
                * classification_evaluation
                * plot_confusion_matrix
        """
        try:
            # Convert to
            if isinstance(thresholds_matrix, np.ndarray):
                thresholds_matrix = thresholds_matrix.tolist()

            if not thresholds_matrix:
                thresholds_matrix = [[]]

            if isinstance(thresholds_matrix, list) and not isinstance(
                    thresholds_matrix[0], list):
                thresholds_matrix = list(thresholds_matrix)

            none_required = False
            for vector in thresholds_matrix:
                if not vector or len(vector) == 0:
                    none_required = True

            if not none_required:
                thresholds_matrix.append(None)

            self.__called_from_perform = True

            self.generate_matrix_meta_data(X,
                                           dataset_name + "/_Extras")

            print("\n\n" + "---" * 10 + f'{dataset_name}' + "---" * 10)
            first_iteration = True

            for pred_name, pred_type in self.__pred_funcs_types.items():

                # Nicer formating
                if not first_iteration:
                    print("\n\n\n")
                first_iteration = False

                for thresholds in thresholds_matrix:

                    print(f"Now running classification on {pred_name}", end='')
                    if pred_type == "Predictions":
                        print()
                        thresholds = None
                    else:
                        if thresholds:
                            print(f" on thresholds:")
                            for i,target_val in enumerate(self.__target_values):
                                try:
                                    print(f"\tTarget Value:{target_val}: Prediction weight: {thresholds[i]}")

                                except IndexError:
                                    raise IndexError("Thresholds must of the same length as the target values!")
                        else:
                            print(" on no thresholds.")

                    if display_visuals:
                        try:
                            print(f"\nShape of the data is {X.shape}")
                        except AttributeError:
                            pass

                    self.classification_metrics(X,
                                                y,
                                                pred_name=pred_name,
                                                dataset_name=dataset_name,
                                                thresholds=thresholds,
                                                ignore_metrics=ignore_metrics,
                                                custom_metrics_dict=custom_metrics_dict,
                                                average_scoring=average_scoring,
                                                display_visuals=display_visuals)

                    self.plot_confusion_matrix(X,
                                               y,
                                               pred_name=pred_name,
                                               dataset_name=dataset_name,
                                               thresholds=thresholds,
                                               normalize=True,
                                               display_visuals=display_visuals)

                    self.plot_confusion_matrix(X,
                                               y,
                                               pred_name=pred_name,
                                               dataset_name=dataset_name,
                                               thresholds=thresholds,
                                               normalize=False,
                                               display_visuals=display_visuals)

                    if pred_type == "Probabilities":
                        self.plot_precision_recall_curve(X,
                                                         y,
                                                         pred_name=pred_name,
                                                         dataset_name=dataset_name,
                                                         thresholds=thresholds,
                                                         display_visuals=display_visuals)
                        self.plot_roc_curve(X,
                                            y,
                                            pred_name=pred_name,
                                            dataset_name=dataset_name,
                                            thresholds=thresholds,
                                            display_visuals=display_visuals)

                        if self.__binary_classifcation:

                            self.plot_lift_curve(X,
                                                 y,
                                                 pred_name=pred_name,
                                                 dataset_name=dataset_name,
                                                 thresholds=thresholds,
                                                 display_visuals=display_visuals)

                            self.plot_ks_statistic(X,
                                                   y,
                                                   pred_name=pred_name,
                                                   dataset_name=dataset_name,
                                                   thresholds=thresholds,
                                                   display_visuals=display_visuals)

                            self.plot_cumulative_gain(X,
                                                      y,
                                                      pred_name=pred_name,
                                                      dataset_name=dataset_name,
                                                      thresholds=thresholds,
                                                      display_visuals=display_visuals)

                    if classification_error_analysis:
                        self.classification_error_analysis(X,
                                                           y,
                                                           pred_name,
                                                           dataset_name,
                                                           thresholds=thresholds,
                                                           display_visuals=False,
                                                           save_file=True,
                                                           aggerate_target=True,
                                                           display_print=False,
                                                           suppress_runtime_errors=True,
                                                           aggregate_target_feature=True,
                                                           selected_features=None,
                                                           extra_tables=True,
                                                           statistical_analysis_on_aggregates=True)

                    if classification_correct_analysis:
                        self.classification_correct_analysis(X,
                                                             y,
                                                             pred_name,
                                                             dataset_name,
                                                             thresholds=thresholds,
                                                             display_visuals=False,
                                                             save_file=True,
                                                             aggerate_target=True,
                                                             display_print=False,
                                                             suppress_runtime_errors=True,
                                                             aggregate_target_feature=True,
                                                             selected_features=None,
                                                             extra_tables=True,
                                                             statistical_analysis_on_aggregates=True)

                    print("-" * (len(dataset_name) + 60) + "\n")
                    if pred_type == "Predictions":
                        break
        finally:
            self.__called_from_perform = False

    def plot_ks_statistic(self,
                          X,
                          y,
                          pred_name,
                          dataset_name,
                          thresholds=None,
                          display_visuals=True,
                          save_file=True,
                          title=None,
                          ax=None,
                          figsize=GRAPH_DEFAULTS.FIGSIZE,
                          title_fontsize='large',
                          text_fontsize='medium'):

        """

            From scikit-plot documentation
            Link: http://tinyurl.com/y3ym5pyc
            Generates the KS Statistic plot from labels and scores/probabilities.

        Args:
            X:
                Feature matrix.

             y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Visualize graph if needed.

            save_file:
                Boolean value to whether or not to save the file.
        """

        filename = f'KS Statistic on {dataset_name} on {self.__model_name}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)
        if not title:
            title = filename

        skplt.metrics.plot_ks_statistic(y,
                                        self.__get_model_probas(pred_name,
                                                                X),
                                        title=title,
                                        ax=ax,
                                        figsize=figsize,
                                        title_fontsize=title_fontsize,
                                        text_fontsize=text_fontsize)
        legend = plt.legend(frameon=1)
        frame = legend.get_frame()
        frame.set_facecolor('white')
        frame.set_edgecolor('white')
        if save_file:
            self.save_plot(filename=filename,
                           sub_dir=sub_dir)

            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")

        if display_visuals:
            plt.show()
        plt.close()

    def plot_roc_curve(self,
                       X,
                       y,
                       pred_name,
                       dataset_name,
                       thresholds=None,
                       display_visuals=True,
                       save_file=True,
                       title=None,
                       ax=None,
                       figsize=GRAPH_DEFAULTS.FIGSIZE,
                       title_fontsize='large',
                       text_fontsize='medium'):

        """

            From scikit-plot documentation
            Link: http://tinyurl.com/y3ym5pyc
            Creates ROC curves from labels and predicted probabilities.

        Args:
             X:
                Feature matrix.

             y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Visualize graph if needed.

            save_file:
                Boolean value to whether or not to save the file.
        """
        filename = f'Roc Curve on {dataset_name} on {self.__model_name}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)
        if not title:
            title = filename

        skplt.metrics.plot_roc(y,
                               self.__get_model_probas(pred_name,
                                                       X),
                               title=title,
                               ax=ax,
                               figsize=figsize,
                               title_fontsize=title_fontsize,
                               text_fontsize=text_fontsize)

        legend = plt.legend(frameon=1)
        frame = legend.get_frame()
        frame.set_facecolor('white')
        frame.set_edgecolor('white')


        if save_file:
            self.save_plot(filename=filename,
                           sub_dir=sub_dir)
            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")

        if display_visuals:
            plt.show()
        plt.close()

    def plot_cumulative_gain(self,
                             X,
                             y,
                             pred_name,
                             dataset_name,
                             thresholds=None,
                             display_visuals=True,
                             save_file=True,
                             title=None,
                             ax=None,
                             figsize=GRAPH_DEFAULTS.FIGSIZE,
                             title_fontsize='large',
                             text_fontsize='medium'):

        """

            From scikit-plot documentation
            Link: http://tinyurl.com/y3ym5pyc
            Plots calibration curves for a set of classifier probability estimates.

        Args:
            X:
                Feature matrix.

             y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Visualize graph if needed.

            save_file:
                Boolean value to whether or not to save the file.
        """
        filename = f'Cumulative Gain gain on {dataset_name} on {self.__model_name}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)
        if not title:
            title = filename

        skplt.metrics.plot_cumulative_gain(y,
                                           self.__get_model_probas(pred_name,
                                                               X),
                                           title=title,
                                           ax=ax,
                                           figsize=figsize,
                                           title_fontsize=title_fontsize,
                                           text_fontsize=text_fontsize)

        legend = plt.legend(frameon=1)
        frame = legend.get_frame()
        frame.set_facecolor('white')
        frame.set_edgecolor('white')


        if save_file:
            self.save_plot(filename=filename,
                           sub_dir=sub_dir)

            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")

        if display_visuals:
            plt.show()
        plt.close()

    def plot_precision_recall_curve(self,
                                    X,
                                    y,
                                    pred_name,
                                    dataset_name,
                                    thresholds=None,
                                    display_visuals=True,
                                    save_file=True,
                                    title=None,
                                    plot_micro=True,
                                    classes_to_plot=None,
                                    ax=None,
                                    figsize=GRAPH_DEFAULTS.FIGSIZE,
                                    cmap='nipy_spectral',
                                    title_fontsize='large',
                                    text_fontsize='medium'):
        """

            From scikit-plot documentation
            Link: http://tinyurl.com/y3ym5pyc
            Plots precision recall curve plot based on the models predictions.

        Args:
            X:
                Feature matrix.

             y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Visualize graph if needed.

            save_file:
                Boolean value to whether or not to save the file.
        """

        filename = f'Precision Recall on {dataset_name} on {self.__model_name}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)
        if not title:
            title = filename

        skplt.metrics.plot_precision_recall(y,
                                            self.__get_model_probas(pred_name,
                                                                    X),
                                            title=title,
                                            plot_micro=plot_micro,
                                            classes_to_plot=classes_to_plot,
                                            ax=ax,
                                            figsize=figsize,
                                            cmap=cmap,
                                            title_fontsize=title_fontsize,
                                            text_fontsize=text_fontsize)

        legend = plt.legend(frameon=1)
        frame = legend.get_frame()
        frame.set_facecolor('white')
        frame.set_edgecolor('white')

        if save_file:
            self.save_plot(filename=filename,
                           sub_dir=sub_dir)

            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")

        if display_visuals:
            plt.show()
        plt.close()

    def plot_lift_curve(self,
                        X,
                        y,
                        pred_name,
                        dataset_name,
                        thresholds=None,
                        display_visuals=True,
                        save_file=True,
                        title=None,
                        ax=None,
                        figsize=GRAPH_DEFAULTS.FIGSIZE,
                        title_fontsize='large',
                        text_fontsize='medium'):
        """

            From scikit-plot documentation
            Link: http://tinyurl.com/y3ym5pyc
            The lift curve is used to determine the effectiveness of a binary classifier.
            A detailed explanation can be found at http://tinyurl.com/csegj9.
            The implementation here works only for binary classification.

        Args:
            X:
                Feature matrix.

             y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Visualize graph if needed.

            save_file:
                Boolean value to whether or not to save the file.
        """

        filename = f'Lift Curve on {dataset_name} on {self.__model_name}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)
        if not title:
            title = filename

        skplt.metrics.plot_lift_curve(y,
                                      self.__get_model_probas(pred_name,
                                                              X),
                                      title=title,
                                      ax=ax,
                                      figsize=figsize,
                                      title_fontsize=title_fontsize,
                                      text_fontsize=text_fontsize)


        legend = plt.legend(frameon=1)
        frame = legend.get_frame()
        frame.set_facecolor('white')
        frame.set_edgecolor('white')

        if save_file:
            self.save_plot(filename=filename,
                           sub_dir=sub_dir)

            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")

        if display_visuals:
            plt.show()
        plt.close()

    def plot_confusion_matrix(self,
                              X,
                              y,
                              pred_name,
                              dataset_name,
                              thresholds=None,
                              display_visuals=True,
                              save_file=True,
                              title=None,
                              normalize=False,
                              hide_zeros=False,
                              hide_counts=False,
                              x_tick_rotation=0,
                              ax=None,
                              figsize=GRAPH_DEFAULTS.FIGSIZE,
                              cmap='Blues',
                              title_fontsize='large',
                              text_fontsize='medium'):
        """

            From scikit-plot documentation
            Link: http://tinyurl.com/y3ym5pyc
            Creates a confusion matrix plot based on the models predictions.

        Args:
            X:
                Feature matrix.

             y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Visualize graph if needed.

            save_file:
                Boolean value to whether or not to save the file.
        """
        filename = f'Confusion Matrix: {dataset_name} on {self.__model_name} Normalized: {normalize}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)
        if not title:
            title = filename

        warnings.filterwarnings('ignore')
        ax = skplt.metrics.plot_confusion_matrix(
             self.__get_model_prediction(pred_name,
                                         X,
                                         thresholds),
             y,
             title=title,
             normalize=normalize,
             hide_zeros=hide_zeros,
             hide_counts=hide_counts,
             x_tick_rotation=x_tick_rotation,
             ax=ax,
             figsize=figsize,
             cmap=cmap,
             title_fontsize=title_fontsize,
             text_fontsize=text_fontsize)
        warnings.filterwarnings('default')

        if save_file:
            self.save_plot(filename=filename,
                           sub_dir=sub_dir)
            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")

        if display_visuals:
            plt.show()
        plt.close()

    def classification_metrics(self,
                               X,
                               y,
                               pred_name,
                               dataset_name,
                               thresholds=None,
                               display_visuals=True,
                               save_file=True,
                               title="",
                               custom_metrics_dict=dict(),
                               ignore_metrics=[],
                               average_scoring=["micro",
                                                "macro",
                                                "weighted"]):
        """

            Creates a dataframe based on the prediction metrics
            of the feature matrix and target vector.

        Args:
            X:
                Feature matrix.

            y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned stored
                in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Display tables.

            save_file:
                Determines whether or not to save the generated document.

            title:
                Adds to the column 'Metric Score'.

            custom_metrics_dict:
                Pass the name of metric(s) and the function definition(s) in a
                dictionary.

            ignore_metrics:
                Specify the default metrics to not apply to the classification
                data_analysis.
                    * Precision
                    * MCC
                    * Recall
                    * F1-Score
                    * Accuracy

            average_scoring:
                Determines the type of averaging performed on the data.
                    * micro
                    * macro
                    * weighted

        Returns:
            Return a dataframe object of the metrics value.
        """
        filename = f'Metric Evaluation on {dataset_name} on {self.__model_name}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)

        if not isinstance(average_scoring, list):
            average_scoring = [average_scoring]

        # Default metric name's and their function
        metric_functions = dict()
        metric_functions["Precision"] = precision_score
        metric_functions["MCC"] = matthews_corrcoef
        metric_functions["Recall"] = recall_score
        metric_functions["F1-Score"] = f1_score
        metric_functions["Accuracy"] = accuracy_score

        warnings.filterwarnings('ignore')

        # Ignore default metrics if needed
        for remove_metric in ignore_metrics:
            if remove_metric in metric_functions:
                del metric_functions[remove_metric]

        # Add in custom metrics
        if len(custom_metrics_dict.keys()):
            metric_functions.update(custom_metrics_dict)

        # Evaluate model on metrics
        evaluation_report = dict()
        for metric_name in metric_functions.keys():
            for average_score in average_scoring:

                model_predictions = self.__get_model_prediction(pred_name,
                                                                X,
                                                                thresholds)
                try:
                    evaluation_report[f'{metric_name}({average_score})'] = \
                        metric_functions[metric_name](y_true=y,
                                                      y_pred=model_predictions,
                                                      average=average_score)
                except TypeError:

                    if metric_name not in evaluation_report.keys():
                        evaluation_report[metric_name] = metric_functions[
                            metric_name](y,
                                         model_predictions)
                    continue

        warnings.filterwarnings('default')

        if title and len(title) > 0:
            index_name = f"Metric Scores ({title})"
        else:
            index_name = "Metric Scores"

        # ---
        evaluation_report = pd.DataFrame({index_name:
                                              [f'{metric_score:.4f}'
                                               for metric_score
                                               in evaluation_report.values()]},
                                         index=list(evaluation_report.keys()))

        if display_visuals:
            if self.__notebook_mode:
                display(evaluation_report)
            else:
                print(evaluation_report)

        if save_file:
            df_to_image(evaluation_report,
                        self.folder_path,
                        sub_dir,
                        convert_to_filename(filename),
                        col_width=20,
                        show_index=True,
                        format_float_pos=4)

            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")


    def classification_correct_analysis(self,
                                        X,
                                        y,
                                        pred_name,
                                        dataset_name,
                                        thresholds=None,
                                        display_visuals=True,
                                        save_file=True,
                                        aggerate_target=False,
                                        display_print=True,
                                        suppress_runtime_errors=True,
                                        aggregate_target_feature=True,
                                        selected_features=None,
                                        extra_tables=True,
                                        statistical_analysis_on_aggregates=True):
        """

            Compares the actual target value to the predicted value and performs
            analysis of all the data.

        Args:
            X: np.matrix or lists of lists
                Feature matrix.

            y: collection object
                Target data vector.

            pred_name: str
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name: str
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            feature_order: collection object
                Features names in proper order to re-create the pandas dataframe.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            extra_tables: bool
                When handling two types of features if set to true this will
                    generate any extra tables that might be helpful.
                    Note -
                        These graphics may create duplicates if you already applied
                        an aggregation in 'perform_analysis'

            aggregate_target_feature: bool
                Aggregate the data of the target feature if the data is
                non-continuous data.

                Note
                    In the future I will have this also working with continuous
                    data.

            selected_features: collection object of features
                Will only focus on these selected feature's and will ignore
                the other given features.

            statistical_analysis_on_aggregates: bool
                If set to true then the function 'statistical_analysis_on_aggregates'
                will run; which aggregates the data of the target feature either
                by discrete values or by binning/labeling continuous data.
        """

        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)

        model_predictions = self.__get_model_prediction(pred_name,
                                                        X,
                                                        thresholds=thresholds)

        if sum(model_predictions != y) == len(y):
            print("Your model predicted everything correctly for this dataset! No correct analysis needed!")
            print("Also sorry for your model...zero correct? Dam...")
        else:
            print("\n\n" + "*" * 10 +
                  "Generating graphs for when the model predicted correctly" +
                  "*" * 10 + "\n")

            # Generate error dataframe
            correct_df = pd.DataFrame.from_records(X[model_predictions == y])
            correct_df.columns = self.__feature_order
            correct_df[self.__target_feature] = y[model_predictions == y]

            # Directory path
            create_dir_structure(self.folder_path,
                                 sub_dir + "/Correctly Predicted Data/All Correct Data")
            output_path = f"{self.folder_path}/{sub_dir}/Correctly Predicted Data"

            tmp_df_features = copy.deepcopy(self.__df_features)

            data_encoder = DataEncoder(create_file=False)

            data_encoder.revert_dummies(correct_df,
                                        tmp_df_features,
                                        qualitative_features=list(self.__df_features.get_dummy_encoded_features().keys()))

            data_encoder.decode_data(correct_df,
                                     tmp_df_features,
                                     apply_value_representation=False)

            data_encoder.apply_value_representation(correct_df,
                                                    tmp_df_features)
            del data_encoder

            # Create feature analysis
            feature_analysis = FeatureAnalysis(tmp_df_features,
                                               overwrite_full_path=output_path + "/All Correct Data")
            feature_analysis.perform_analysis(correct_df,
                                              dataset_name=dataset_name,
                                              target_features=[self.__target_feature],
                                              save_file=save_file,
                                              selected_features=selected_features,
                                              suppress_runtime_errors=suppress_runtime_errors,
                                              display_print=display_print,
                                              display_visuals=display_visuals,
                                              dataframe_snapshot=False,
                                              aggregate_target_feature=aggregate_target_feature,
                                              extra_tables=extra_tables,
                                              statistical_analysis_on_aggregates=statistical_analysis_on_aggregates)

            # Aggregate target by predicted and actual
            if aggerate_target:
                targets = set(y)
                for pred_target in targets:
                    if pred_target != pred_target:
                        create_dir_structure(output_path,
                                             f"/Actual and Predicted:{pred_target}")

                        # Create predicted vs actual dataframe
                        tmp_df = copy.deepcopy(correct_df[correct_df[
                                                              self.__target_feature] == pred_target])

                        if tmp_df.shape[0]:
                            # Create feature analysis directory structure with given graphics
                            feature_analysis = FeatureAnalysis(
                                tmp_df_features,
                                overwrite_full_path=f"/Actual and Predicted:{pred_target}")
                            feature_analysis.perform_analysis(tmp_df,
                                                              dataset_name=dataset_name,
                                                              target_features=[
                                                                  self.__target_feature],
                                                              save_file=save_file,
                                                              selected_features=selected_features,
                                                              suppress_runtime_errors=suppress_runtime_errors,
                                                              display_print=display_print,
                                                              display_visuals=display_visuals,
                                                              dataframe_snapshot=False,
                                                              extra_tables=extra_tables,
                                                              statistical_analysis_on_aggregates=statistical_analysis_on_aggregates)

    def classification_error_analysis(self,
                                      X,
                                      y,
                                      pred_name,
                                      dataset_name,
                                      thresholds=None,
                                      display_visuals=True,
                                      save_file=True,
                                      aggerate_target=False,
                                      display_print=True,
                                      suppress_runtime_errors=True,
                                      aggregate_target_feature=True,
                                      selected_features=None,
                                      extra_tables=True,
                                      statistical_analysis_on_aggregates=True):
        """

            Compares the actual target value to the predicted value and performs
            analysis of all the data.

        Args:
            X: np.matrix or lists of lists
                Feature matrix.

            y: collection object
                Target data vector.

            pred_name: str
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name: str
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            feature_order: collection object
                Features names in proper order to re-create the pandas dataframe.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals: bool
                Boolean value to whether or not to display visualizations.

            display_print: bool
                Determines whether or not to print function's embedded print
                statements.

            save_file: bool
                Boolean value to whether or not to save the file.

            dataframe_snapshot: bool
                Boolean value to determine whether or not generate and compare a
                snapshot of the dataframe in the dataset's directory structure.
                Helps ensure that data generated in that directory is correctly
                associated to a dataframe.

            suppress_runtime_errors: bool
                If set to true; when generating any graphs will suppress any runtime
                errors so the program can keep running.

            extra_tables: bool
                When handling two types of features if set to true this will
                    generate any extra tables that might be helpful.
                    Note -
                        These graphics may create duplicates if you already applied
                        an aggregation in 'perform_analysis'

            aggregate_target_feature: bool
                Aggregate the data of the target feature if the data is
                non-continuous data.

                Note
                    In the future I will have this also working with continuous
                    data.

            selected_features: collection object of features
                Will only focus on these selected feature's and will ignore
                the other given features.

            statistical_analysis_on_aggregates: bool
                If set to true then the function 'statistical_analysis_on_aggregates'
                will run; which aggregates the data of the target feature either
                by discrete values or by binning/labeling continuous data.
        """

        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)

        model_predictions = self.__get_model_prediction(pred_name,
                                                        X,
                                                        thresholds=thresholds)

        if sum(model_predictions == y) == len(y):
            print("Your model predicted everything correctly for this dataset! No error analysis needed!")
        else:
            print("\n\n" + "*" * 10 +
                  "Generating graphs for when the model predicted incorrectly" +
                  "*" * 10 + "\n")

            # Generate error dataframe
            error_df = pd.DataFrame.from_records(X[model_predictions != y])
            error_df.columns = self.__feature_order
            error_df[self.__target_feature] = y[model_predictions != y]
            error_df.reset_index(drop=True,
                                 inplace=True)

            # Directory path
            create_dir_structure(self.folder_path,
                                 sub_dir + "/Incorrectly Predicted Data/All Incorrect Data")
            output_path = f"{self.folder_path}/{sub_dir}/Incorrectly Predicted Data"

            tmp_df_features = copy.deepcopy(self.__df_features)

            data_encoder = DataEncoder(create_file=False)

            data_encoder.revert_dummies(error_df,
                                        tmp_df_features,
                                        qualitative_features=list(self.__df_features.get_dummy_encoded_features().keys()))

            data_encoder.decode_data(error_df,
                                     tmp_df_features,
                                     apply_value_representation=False)

            data_encoder.apply_value_representation(error_df,
                                                    tmp_df_features)
            del data_encoder

            # Create feature analysis
            feature_analysis = FeatureAnalysis(tmp_df_features,
                                               overwrite_full_path=output_path + "/All Incorrect Data")
            feature_analysis.perform_analysis(error_df,
                                              dataset_name=dataset_name,
                                              target_features=[self.__target_feature],
                                              save_file=save_file,
                                              selected_features=selected_features,
                                              suppress_runtime_errors=suppress_runtime_errors,
                                              display_print=display_print,
                                              display_visuals=display_visuals,
                                              dataframe_snapshot=False,
                                              aggregate_target_feature=aggregate_target_feature,
                                              extra_tables=extra_tables,
                                              statistical_analysis_on_aggregates=statistical_analysis_on_aggregates)

            # Aggregate target by predicted and actual
            if aggerate_target:
                targets = set(y)
                prediction_feature = self.__target_feature + "_MODEL_PREDICTIONS_"
                error_df[prediction_feature] = model_predictions[model_predictions != y]
                for actual_target in targets:
                    for pred_target in targets:
                        if pred_target != actual_target:
                            create_dir_structure(output_path,
                                                 f"/Predicted:{pred_target} Actual: {actual_target}")

                            # Create predicted vs actual dataframe
                            tmp_df = copy.deepcopy(error_df[error_df[
                                                                self.__target_feature] == actual_target][
                                                       error_df[
                                                           prediction_feature] == pred_target])

                            tmp_df.drop(columns=[prediction_feature],
                                        inplace=True)
                            if tmp_df.shape[0]:
                                # Create feature analysis directory structure with given graphics
                                feature_analysis = FeatureAnalysis(tmp_df_features,
                                                                   overwrite_full_path=f"{output_path}/Predicted:{pred_target} Actual: {actual_target}")
                                feature_analysis.perform_analysis(tmp_df,
                                                                  dataset_name=dataset_name,
                                                                  target_features=[
                                                                      self.__target_feature],
                                                                  save_file=save_file,
                                                                  selected_features=selected_features,
                                                                  suppress_runtime_errors=suppress_runtime_errors,
                                                                  display_print=display_print,
                                                                  display_visuals=display_visuals,
                                                                  dataframe_snapshot=False,
                                                                  extra_tables=extra_tables,
                                                                  statistical_analysis_on_aggregates=statistical_analysis_on_aggregates)


    def classification_report(self,
                              X,
                              y,
                              pred_name,
                              dataset_name,
                              thresholds=None,
                              display_visuals=True,
                              save_file=True):
        """

            Creates a report of all target's metric evaluations
            based on the model's prediction output from the classification report
            from the sklearn.

        Args:
            X:
                Feature matrix.

             y:
                Target data vector.

            pred_name:
                The name of the prediction function in questioned
                stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

            display_visuals:
                Visualize graph if needed.

            save_file:
                Boolean value to whether or not to save the file.
        """
        filename = f'Classification Report {dataset_name} on {self.__model_name}'
        sub_dir = self.__create_sub_dir_with_thresholds(pred_name,
                                                        dataset_name,
                                                        thresholds)

        # Create dataframe report
        report_df = pd.DataFrame(classification_report(y,
                                                       self.__get_model_prediction(
                                                           pred_name,
                                                           X,
                                                           thresholds),
                                                       output_dict=True))

        # ---
        if display_visuals:
            if self.__notebook_mode:
                display(report_df)
            else:
                print(report_df)

        if save_file:
            # Output dataframe as png
            df_to_image(report_df,
                        self.folder_path,
                        sub_dir,
                        filename,
                        col_width=20,
                        show_index=True,
                        format_float_pos=4)

            if not self.__called_from_perform:
                self.generate_matrix_meta_data(X,
                                               dataset_name + "/_Extras")

    def graph_model_importances(self,
                                feature_order,
                                feature_importances,
                                display_visuals=True):

        feature_importances, feature_order = zip(
            *sorted(zip(feature_importances, feature_order), reverse=True))
        feature_order = list(feature_order)

        plt.figure(figsize=(20, 10))

        palette = "PuBu"

        # Color ranking
        rank_list = np.argsort(-np.array(feature_importances)).argsort()
        pal = sns.color_palette(palette, len(feature_importances))
        palette = np.array(pal[::-1])[rank_list]

        plt.clf()

        plt.title("Feature Importances")

        ax = sns.barplot(x=feature_importances,
                         y=feature_order,
                         palette=palette,
                         order=feature_order)

        self.save_plot("Feature Importances",
                       "_Extras")

        pickle_object_to_file(dict(zip(feature_order, feature_importances)),
                              self.folder_path + "/_Extras",
                              "Feature Importances")

        if self.__notebook_mode and display_visuals:
            plt.show()

        plt.close("all")



    def __get_model_prediction(self,
                               pred_name,
                               X,
                               thresholds=None):
        """

            Finds the model's predicted labels.

        Args:
            X:
                Feature matrix.

            pred_name:
                The name of the prediction function in questioned stored in 'self.__pred_funcs_dict'

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

        Returns:
            Returns back a predicted value based for a given matrix.
            Handles prediction function 'types' Predictions and Probabilities.
            Helps streamline the entire process of evaluating classes.
        """

        # Check if function name exists in the dictionary
        if pred_name not in self.__pred_funcs_types:
            raise KeyError("The function name has not be recorded!")

        # Must be a prediction function
        if self.__pred_funcs_types[pred_name] == "Predictions":
            return self.__pred_funcs_dict[pred_name](X)

        # Output must be continuous; Probabilities
        elif self.__pred_funcs_types[pred_name] == "Probabilities":

            # Validate probabilities
            if thresholds:
                if isinstance(thresholds, list) or \
                        isinstance(thresholds, np.ndarray):
                    if len(thresholds) != len(self.__target_values):
                        raise UnsatisfiedRequirments("Length of thresholds must match the same length as the associated classes.")
                else:
                    raise UnsatisfiedRequirments("Threshold object is not a list or numpy array!")

            # Get model probability output
            model_output = self.__get_model_probas(pred_name,
                                                   X)

            # No thresholds found
            if not thresholds:
                return np.asarray([self.__target_values[np.argmax(proba)]
                                   for proba in model_output])

            # Apply thresholds to model's probability output
            else:
                model_output = model_output - thresholds

                return np.asarray([self.__target_values[np.argmax(proba)]
                                   for proba in model_output])
        else:
            raise ValueError(f"Unknown type '{self.__pred_funcs_types[pred_name]}' was found!")

    def __get_model_probas(self,
                           pred_name,
                           X):
        """

            Attempts to get the probabilities from the prediction function.

        Args:
            X:
                Feature matrix.

            pred_name:
                The name of the prediction function in questioned stored in 'self.__pred_funcs_dict'

        Raises:
             If probabilities isn't possible with the given function that it will invoke an error.

        Returns:
            Returns back a series of values between 0-1 to represent it's confidence.
        """

        if self.__pred_funcs_types[pred_name] == "Probabilities":
            model_output = self.__pred_funcs_dict[pred_name](X)

            # ---
            if isinstance(model_output, list):
                model_output = np.asarray(model_output)

            return model_output
        else:
            raise ProbasNotPossible


    def __create_sub_dir_with_thresholds(self,
                                         pred_name,
                                         dataset_name,
                                         thresholds):
        """

            Iterates through directory structure looking at each text file
            containing a string representation of the given threshold;
            keeps comparing the passed value of 'thresholds' to the text file.

        Args:
            pred_name:
                The name of the prediction function in questioned stored in 'self.__pred_funcs_dict'

            dataset_name:
                The dataset's name; this will create a sub-directory in which your
                generated graph will be inner-nested in.

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

        Returns:
            Looking at the root of the starting directory and looking at each
            '_Thresholds.txt' file to determine if the files can be outputed
            to that directory. The content of the file must match the content
            of the list/numpy array 'thresholds'.
        """

        sub_dir = f'{dataset_name}/{pred_name}'

        # Only generate extra folder structure if function type is Probabilities
        if self.__pred_funcs_types[pred_name] == "Probabilities":

            # ------
            if not thresholds:
                sub_dir = f'{sub_dir}/No Thresholds'
            else:
                i = 0
                sub_dir = f'{sub_dir}/Thresholds'
                tmp_sub_dir = copy.deepcopy(sub_dir)
                while True:
                    threshold_dir = self.folder_path
                    if i > 0:
                        tmp_sub_dir = (sub_dir + f' {i}')
                    threshold_dir += tmp_sub_dir

                    # If file exists with the same thresholds; than use this directory
                    if os.path.exists(threshold_dir):
                        if self.__compare_thresholds_to_saved_thresholds(
                                threshold_dir,
                                thresholds):
                            sub_dir = tmp_sub_dir
                            break

                    # Create new directory
                    else:
                        os.makedirs(threshold_dir)
                        write_object_text_to_file(thresholds,
                                                  threshold_dir,
                                                  "_Thresholds")
                        sub_dir = tmp_sub_dir
                        break

                    # Iterate for directory name change
                    i += 1

        return sub_dir

    def __compare_thresholds_to_saved_thresholds(self,
                                                 directory_path,
                                                 thresholds):
        """

            Compare the thresholds object to a threshold text file found in
            the directory; returns true if the file exists and the object's
            value matches up.

        Args:
            directory_path:
                Path to the given folder where the "_Thresholds.txt"

            thresholds:
                If the model outputs a probability list/numpy array then we apply
                thresholds to the ouput of the model.
                For classification only; will not affect the direct output of
                the probabilities.

        Returns:
            Compare the thresholds object to the text file; returns true if
            the file exists and the object's value matches up.
        """

        file_directory = correct_directory_path(directory_path)

        if os.path.exists(file_directory):

            # Extract file contents and convert to a list object
            file = open(file_directory + "_Thresholds.txt", "r")
            line = file.read()
            converted_list = line.split("=")[-1].strip().strip('][').split(
                ', ')
            converted_list = [float(val) for val in converted_list]
            file.close()

            if thresholds == converted_list:
                return True
            else:
                return False
        else:
            return False


# Import libs
import os


# Taken from utils.sys_utils
def get_all_directories_from_path(directory_path):
    """
    directory_path:
        Given path that already exists.

    Returns:
        Returns back a set a directories with the provided path.
    """

    dirs_in_paths = []
    for (dirpath, dirnames, filenames) in os.walk(directory_path):
        dirs_in_paths.extend(dirnames)
        break

    return set(dirs_in_paths)


def get_all_files_from_path(directory_path,
                            file_extension=None):
    """
    directory_path:
        Given path that already exists.

    file_extension:
        Only return files that have a given extension.

    Returns:
        Returns back a set a filenames with the provided path.
    """

    files_in_paths = []
    for (dirpath, dirnames, filenames) in os.walk(directory_path):

        if file_extension:
            file_extension = file_extension.replace(".","")
            for file in filenames:
                if file.endswith(f'.{file_extension}'):
                    files_in_paths.append(file)
        else:
            files_in_paths.extend(filenames)
        break

    return set(files_in_paths)

# Get the current working directory
current_work_dir = os.getcwd()
project_dir = current_work_dir  + "/eflow/"
print(project_dir)
# Get all directories from project
all_dirs = get_all_directories_from_path(project_dir)
print(all_dirs)
for dir_name in all_dirs:

    # Ignore any hidden files
    if dir_name[0] == "_":
        continue

    # Ignore utils for now
    if dir_name == "utils":
        continue

    dir_files = get_all_files_from_path(project_dir + dir_name,
                                        "py")
    print(dir_files)
    i = 0
    for file_name in dir_files:

        print(file_name)
        # Ignore hidden files
        if file_name[0] == "_":
            continue

        with open(f'{project_dir}{dir_name}/{file_name}') as fp:

            function_def = ""
            line = fp.readline()
            while line:
                if ("def " in line and "def _" not in line):
                    def_start = True
                    function_def += line.split("def ")[1].replace("(self,","(").strip()
                    while line:
                        line = fp.readline()
                        
                        if "):" in line:
                            def_start = False
                            break
                        else:
                            line = line.strip()
                            function_def += line
                    function_def += ")"
                    break
                line = fp.readline()
            
            print(function_def)
            function_doc = ""
            doc_found = False
            while line:
                print(line)
                if "\"\"\"" in line:
                    doc_found != doc_found
                elif doc_found:
                    function_doc += line

                line = fp.readline()

            print(function_doc)
                
#        print(function_def)
        if i == 2:
            break
        else:
            i += 1
    break



from math import log, e
import numpy as np
import math

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


def get_unbalanced_threshold(target_amount,
                             max_binary_threshold=.65):
    max_unbalanced_class_threshold = (1 / target_amount) / ((1/2)/(max_binary_threshold-(1/2))) + (
                1 / target_amount)
    min_unbalanced_class_threshold = (1 - max_unbalanced_class_threshold) / (
                target_amount - 1)

    return max_unbalanced_class_threshold, min_unbalanced_class_threshold


def calculate_entropy(labels,
                      base=None):
    n_labels = len(labels)

    if n_labels <= 1:
        return 0

    value, counts = np.unique(labels, return_counts=True)
    probs = counts / n_labels
    n_classes = np.count_nonzero(probs)

    if n_classes <= 1:
        return 0

    ent = 0.

    # Compute entropy
    base = e if base is None else base
    for i in probs:
        ent -= i * log(i, base)

    return ent


def euclidean_distance(v1, v2):
    dist = [(a - b)**2 for a, b in zip(v1, v2)]
    dist = math.sqrt(sum(dist))
    return dist


from eflow._hidden.parent_objects import DataPipelineSegment

import copy

__author__ = "Eric Cacciavillani"
__copyright__ = "Copyright 2019, eFlow"
__credits__ = ["Eric Cacciavillani"]
__license__ = "MIT"
__maintainer__ = "EricCacciavillani"
__email__ = "eric.cacciavillani@gmail.com"


class TypeFixer(DataPipelineSegment):
    """
        Attempts to convert features to the correct types. Will update the
        dataframe and df_features.
    """
    def __init__(self,
                 segment_id=None,
                 create_file=True):
        """
        Args:
            segment_id:
                Reference id to past segments of this object.

        Note/Caveats:
            When creating any public function that will be part of the pipeline's
            structure it is important to follow this given template. Also,
            try not to use _add_to_que. Can ruin the entire purpose of this
            project.
        """
        DataPipelineSegment.__init__(self,
                                     object_type=self.__class__.__name__,
                                     segment_id=segment_id,
                                     create_file=create_file)

    def convert_types(self,
                      df,
                      df_features,
                      user_logic=True):
        pass


# Extended subclass of LabelEncoder
class MyEncoder(preprocessing.LabelEncoder):
    
    def __init__(self,input_labels):
        
        # Error check parameter variable
        if not isinstance(input_labels, (list, tuple, np.ndarray)):
            exit("You sent a object that is not an array type!")
        
        # Create label encoder and fit the labels
        self.fit(input_labels)
        
        # Creating dict
        self.__encode_label = defaultdict(dict)
        self.__encode_num = defaultdict(dict)
        
        # Init dict
        for i, item in enumerate(self.classes_):
            self.__encode_label[item] = i
            self.__encode_num[i] = item

    # Print the label to number relationship
    def displayEncodedMap(self):
        
        print("\nLabel mapping:\n")
        for i, item in enumerate(self.__encoder.classes_):
            print("\t",item, '-->', i)
            
            # Draw a line
            print("-"*30,"\n")
        
        # Decode the label to the num associated with it
        def decodeLabel(self,labels):
            if (not isinstance(labels,str)) and ( not isinstance(labels, (list, tuple, np.ndarray))):
                exit("You sent value(s) that are not label(s)!")
        
            if isinstance(labels,str):
                
                if labels in self.__encode_label:
                    return self.__encode_label[labels]
                else:
                    exit("This key does not exist in the encoder!")

            else:
                return list(self.transform(labels))
        
        # Decode the num to the label associated with it
        def decodeNum(self,values):
            if (not isinstance(values,int)) and (not isinstance(values, (list, tuple, np.ndarray))):
                exit("You sent value(s) that are not integer(s)!")
        
            if isinstance(values,int):
                
                if values in self.__encode_num:
                    return self.__encode_num[values]
                else:
                    exit("This key does not exist in the encoder!")

            else:
                return list(self.inverse_transform(values))
